import inspect
import os
import textwrap
import typing

from civis.resources import generate_classes_maybe_cached
from civis.response import Response


CLIENT_PYI_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
    "client.pyi",
)


def _get_endpoint_class_name(endpoint_name):
    # Factor out this helper function for consistency.
    return f"_{endpoint_name.title()}"


def _get_annotation(param):
    if param.name == "self":
        return ""
    elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
        return param.annotation
    else:
        return f"{param.annotation} | None"


def _extract_nested_response_classes(response_classes, return_type):
    response_classes[return_type.__name__] = return_type
    for typ in return_type.__annotations__.values():
        if isinstance(typ, str):
            continue
        if isinstance(typ, typing._GenericAlias):
            typ = typing.get_args(typ)[0]
        response_classes = _extract_nested_response_classes(response_classes, typ)
    return response_classes


def generate_client_pyi(client_pyi_path, api_spec_path):
    classes = generate_classes_maybe_cached(
        api_spec_path, api_key="not_needed", api_version="1.0"
    )

    with open(client_pyi_path, "w") as f:
        f.write(
            """# This file is auto-generated by tools/update_civis_api_spec.py.
# Do not edit it by hand.

from collections import OrderedDict
from collections.abc import Iterator
from typing import Any, List

from civis.response import Response

"""
        )

        response_classes = {}

        for endpoint_name, endpoint_class in classes.items():
            f.write(f"class {_get_endpoint_class_name(endpoint_name)}:\n")
            method_defs = []
            for method_name, method in vars(endpoint_class).items():
                method_def = ""
                if method_name.startswith("_"):
                    continue
                signature = inspect.signature(method)
                return_type = signature.return_annotation
                if return_type is not Response:
                    if return_type.__name__ == "Iterator":
                        response_classes = _extract_nested_response_classes(
                            response_classes, typing.get_args(return_type)[0]
                        )
                    else:
                        response_classes = _extract_nested_response_classes(
                            response_classes, return_type
                        )
                params = inspect.signature(method).parameters
                method_def += f"    def {method_name}(\n"
                asterisk_added = False
                for param_name, param in params.items():
                    annotation = _get_annotation(param)
                    if param_name == "self":
                        method_def += "        self,\n"
                    elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                        method_def += f"        {param_name}: {annotation},\n"
                    else:
                        if (
                            not asterisk_added
                            and param.kind == inspect.Parameter.KEYWORD_ONLY
                        ):
                            method_def += "        *,\n"
                            asterisk_added = True
                        method_def += f"        {param_name}: {annotation} = ...,\n"
                if return_type.__name__ == "Iterator":
                    return_str = f"Iterator[{typing.get_args(return_type)[0].__name__}]"
                elif method_name.startswith("list"):
                    return_str = f"List[{return_type.__name__}]"
                else:
                    return_str = return_type.__name__
                method_def += f"    ) -> {return_str}:\n"
                method_doc = textwrap.indent(method.__doc__, " " * 8).lstrip()
                method_def += f'        """{method_doc}\n        """\n        ...\n'
                method_defs.append(method_def)
            f.write("\n".join(method_defs))
            f.write("\n")

        for response_class in response_classes.values():
            if len(line1 := f"class {response_class.__name__}(Response):") <= 88:
                f.write(f"{line1}\n")
            elif len(line2 := f"class {response_class.__name__}(") <= 88:
                f.write(f"{line2}\n    Response\n):\n")
            else:
                f.write(
                    f"class {response_class.__name__}(  # noqa: E501\n"
                    "    Response\n):\n"
                )
            for name, anno in response_class.__annotations__.items():
                inner_anno_str = None
                if isinstance(anno, str):
                    anno_str = anno
                elif isinstance(anno, typing._GenericAlias):
                    inner_anno_str = typing.get_args(anno)[0].__name__
                    anno_str = f"{anno.__name__}[{inner_anno_str}]"
                else:
                    anno_str = anno.__name__
                if len(line := f"    {name}: {anno_str}") <= 88:
                    f.write(f"{line}\n")
                elif anno_str.startswith("List"):
                    f.write(f"    {name}: List[\n        {inner_anno_str}\n    ]\n")
                else:
                    f.write(f"    {name}: (\n        {anno_str}\n    )\n")
            f.write("\n")

        f.write(
            """# Need the individual endpoint classes defined first as above,
# before we can define APIClient to use them.
class APIClient:
    default_credential: int | None
    default_database_credential_id: int | None
    username: str
    feature_flags: tuple[str]
    last_response: Any
    def __init__(
        self,
        api_key: str | None = ...,
        return_type: str = ...,
        api_version: str = ...,
        local_api_spec: OrderedDict | str | None = ...,
        force_refresh_api_spec: bool = ...,
    ): ...
    def get_aws_credential_id(
        self,
        cred_name: str | int,
        owner: str | None = None,
    ) -> int: ...
    def get_database_credential_id(
        self,
        username: str | int,
        database_name: str | int,
    ) -> int: ...
    def get_database_id(
        self,
        database: str | int,
    ) -> int: ...
    def get_storage_host_id(
        self,
        storage_host: str | int,
    ) -> int: ...
    def get_table_id(
        self,
        table: str,
        database: str | int,
    ) -> int: ...
"""
        )
        for endpoint_name in classes:
            f.write(
                f"    {endpoint_name} = {_get_endpoint_class_name(endpoint_name)}()\n"  # noqa: E501
            )



# The `camel_to_snake` function is used in multiple modules.
# To avoid creating import overhead, it's defined in a separate module here
# as opposed to in a module that itself has a fair amount of overhead.

import re


UNDERSCORER1 = re.compile(r"(.)([A-Z][a-z]+)")
UNDERSCORER2 = re.compile("([a-z0-9])([A-Z])")


def camel_to_snake(word):
    # https://gist.github.com/jaytaylor/3660565
    word = UNDERSCORER1.sub(r"\1_\2", word)
    return UNDERSCORER2.sub(r"\1_\2", word).lower()



from collections import OrderedDict
import json
from unittest import mock

import requests

from civis.base import CivisAPIError
from civis.service_client import (
    ServiceClient,
    ServiceEndpoint,
    _get_service,
    _parse_service_path,
    parse_service_api_spec,
    to_camelcase,
)
import pytest

MOCK_SERVICE_ID = 0

MOCK_URL = "www.survey-url.com"


@pytest.fixture
def mock_swagger():
    return {
        "info": {"title": "Test API Client", "version": "1.0"},
        "paths": {
            "/some-resources": {
                "get": {
                    "description": "",
                    "responses": {
                        "200": {
                            "description": "Returns a list",
                        }
                    },
                    "summary": "List Resources",
                    "tags": ["tag"],
                }
            },
            "/some-resources/{id}": {
                "get": {
                    "description": "",
                    "responses": {
                        "200": {
                            "description": "Returns a Resource",
                        }
                    },
                    "summary": "Get Resources",
                    "tags": ["tag"],
                },
                "patch": {
                    "description": "",
                    "parameters": [
                        {
                            "description": "The id of the Resource",
                            "in": "path",
                            "name": "id",
                            "required": True,
                            "type": "integer",
                        },
                        {
                            "description": "The fields and values to edit",
                            "in": "body",
                            "name": "body",
                            "required": True,
                            "schema": {
                                "properties": {
                                    "field": {
                                        "description": "a property value",
                                        "type": "string",
                                    }
                                }
                            },
                        },
                    ],
                    "responses": {
                        "200": {
                            "description": "Edits Resource",
                        }
                    },
                    "summary": "Patch Resources",
                    "tags": ["tag"],
                },
            },
        },
        "swagger": "2.0",
    }


@pytest.fixture
def mock_operations(mock_swagger):
    ops_json = mock_swagger["paths"]["/some-resources"]
    mock_ops_str = str(json.dumps(ops_json))
    mock_operations = json.JSONDecoder(object_pairs_hook=OrderedDict).decode(
        mock_ops_str
    )  # noqa: E501
    return mock_operations


@mock.patch("civis.service_client.ServiceClient.generate_classes_maybe_cached")
@mock.patch("civis.service_client.ServiceClient.get_base_url")
def test_service_client(url_mock, classes_mock):
    url_mock.return_value = MOCK_URL
    classes_mock.return_value = {}

    sc = ServiceClient(MOCK_SERVICE_ID)

    spec_endpoint = "/endpoints"

    assert sc._api_key is None
    assert sc._service_id == MOCK_SERVICE_ID
    assert sc._base_url == MOCK_URL
    assert sc._root_path is None
    assert sc._swagger_path == spec_endpoint

    # Custom root path
    sc = ServiceClient(MOCK_SERVICE_ID, root_path="/api")
    assert sc._root_path == "/api"

    # Custom Swagger path
    sc = ServiceClient(MOCK_SERVICE_ID, swagger_path="/spec")
    assert sc._swagger_path == "/spec"

    # Passed in API Key
    sc = ServiceClient(MOCK_SERVICE_ID, api_key="this_is_an_API_key")
    assert sc._api_key == "this_is_an_API_key"


def test_service_endpoint():
    service_client_mock = mock.Mock()
    se = ServiceEndpoint(service_client_mock)

    assert se._return_type == "civis"
    assert se._client == service_client_mock


def test_parse_service_path(mock_operations):
    mock_path = "/some-resource/sub-resource/{id}"
    base_path, methods = _parse_service_path(mock_path, mock_operations)

    assert base_path == "some_resource"
    assert "get_sub_resource" in methods[0]

    mock_path = "/some-resource/{id}"
    base_path, methods = _parse_service_path(mock_path, mock_operations)

    assert base_path == "some_resource"
    assert "get" in methods[0]


def test_parse_path__with_root(mock_operations):
    root_path = "/some-resource"

    mock_path = "/some-resource/sub-resource/{id}"
    base_path, methods = _parse_service_path(
        mock_path, mock_operations, root_path=root_path
    )

    assert base_path == "sub_resource"
    assert "get" in methods[0]


def test_parse_service_api_spec(mock_swagger):
    classes = parse_service_api_spec(mock_swagger)
    assert "some_resources" in classes


@mock.patch("civis.service_client.requests.Session.get")
@mock.patch("civis.service_client.auth_service_session")
@mock.patch("civis.service_client.ServiceClient.generate_classes_maybe_cached")
@mock.patch("civis.service_client.ServiceClient.get_base_url")
def test_get_api_spec(
    url_mock, classes_mock, auth_session_mock, mock_response, mock_swagger
):
    mock_response.return_value = mock.Mock(ok=True)
    mock_response.return_value.json.return_value = mock_swagger

    url_mock.return_value = MOCK_URL
    classes_mock.return_value = {}

    sc = ServiceClient(MOCK_SERVICE_ID)

    spec = sc.get_api_spec()
    assert spec == mock_swagger


@mock.patch("civis.service_client.parse_service_api_spec")
@mock.patch("civis.service_client.ServiceClient.get_api_spec")
@mock.patch("civis.service_client.ServiceClient.get_base_url")
def test_generate_classes(url_mock, api_spec_mock, parse_mock, mock_swagger):
    api_spec_mock.return_value = {}

    def mock_class_function(client, return_type):
        return "return"

    parse_mock.return_value = {"class": mock_class_function}
    url_mock.return_value = MOCK_URL

    sc = ServiceClient(MOCK_SERVICE_ID, root_path="/foo")

    classes = sc.generate_classes()
    parse_mock.assert_called_once_with(api_spec_mock.return_value, root_path="/foo")

    assert "class" in classes


@mock.patch("civis.service_client.parse_service_api_spec")
@mock.patch("civis.service_client.ServiceClient.get_api_spec")
@mock.patch("civis.service_client.ServiceClient.get_base_url")
def test_generate_classes_maybe_cached(
    url_mock, api_spec_mock, parse_mock, mock_swagger
):
    api_spec_mock.return_value = {}

    def mock_class_function(client, return_type):
        return "return"

    parse_mock.return_value = {"class": mock_class_function}
    url_mock.return_value = MOCK_URL

    sc = ServiceClient(MOCK_SERVICE_ID, root_path="/foo")

    mock_spec_str = str(json.dumps(mock_swagger))
    mock_spec = json.JSONDecoder(object_pairs_hook=OrderedDict).decode(
        mock_spec_str
    )  # noqa: E501
    classes = sc.generate_classes_maybe_cached(mock_spec)

    parse_mock.assert_has_calls(
        [
            # the call from generate_classes_maybe_cached in ServiceClient.__init__
            mock.call({}, root_path="/foo"),
            # the call from generate_classes_maybe_cached in this test
            mock.call(mock_spec, root_path="/foo"),
        ]
    )

    assert "class" in classes


@mock.patch("civis.service_client.ServiceClient.generate_classes_maybe_cached")
@mock.patch("civis.service_client._get_service")
def test_get_base_url(get_service_mock, classes_mock):
    get_service_mock.return_value = {"current_url": MOCK_URL}
    classes_mock.return_value = {}

    sc = ServiceClient(MOCK_SERVICE_ID)

    assert sc._base_url == MOCK_URL
    get_service_mock.assert_called_once_with(sc)


@mock.patch("civis.service_client.ServiceClient.generate_classes_maybe_cached")
@mock.patch("civis.service_client.APIClient")
def test_get_service(mock_client, classes_mock):
    classes_mock.return_value = {}
    sc = ServiceClient(MOCK_SERVICE_ID)
    expected_service = {"current_url": MOCK_URL}
    mock_client.return_value.services.get.return_value = expected_service
    service = _get_service(sc)
    assert service == expected_service


@mock.patch("civis.service_client.ServiceClient.generate_classes_maybe_cached")
@mock.patch("civis.service_client.APIClient")
def test_get_service__not_found(mock_client, classes_mock):
    classes_mock.return_value = {}
    sc = ServiceClient(MOCK_SERVICE_ID)
    err_resp = requests.Response()
    err_resp._content = json.dumps(
        {
            "status_code": 404,
            "error": "not_found",
            "errorDescription": "The requested resource could not be found.",
            "content": True,
        }
    ).encode()

    mock_client.return_value.services.get.side_effect = CivisAPIError(err_resp)

    with pytest.raises(CivisAPIError) as excinfo:
        _get_service(sc)

    expected_error = "The requested resource could not be found."
    assert str(excinfo.value) == expected_error


def test_build_path():
    service_client_mock = mock.Mock(_base_url="www.service_url.com", _root_path=None)
    se = ServiceEndpoint(service_client_mock)
    path = se._build_path("/resources")

    assert path == "www.service_url.com/resources"


def test_build_path__with_root():
    service_client_mock = mock.Mock(_base_url="www.service_url.com", _root_path="/api")
    se = ServiceEndpoint(service_client_mock)
    path = se._build_path("/resources")

    assert path == "www.service_url.com/api/resources"


@mock.patch("civis.service_client.requests.Session.request")
@mock.patch("civis.service_client.auth_service_session")
def test_make_request(auth_mock, request_mock):
    service_client_mock = mock.Mock(_base_url="www.service_url.com")
    se = ServiceEndpoint(service_client_mock)

    expected_value = [
        {"id": 1, "url": "www.survey_url.com/1"},
        {"id": 2, "url": "www.survey_url.com/2"},
    ]

    request_mock.return_value = mock.Mock(ok=True)
    request_mock.return_value.json = expected_value

    response = se._make_request("get", "resources/resources")

    assert response.json == expected_value


def test_tocamlecase():
    test_cases = [
        ("snake_case", "SnakeCase"),
        ("Snake_Case", "SnakeCase"),
        ("snakecase", "Snakecase"),
    ]
    for in_word, out_word in test_cases:
        assert to_camelcase(in_word) == out_word



"""Schemas for Civis Platform workflow definitions.

Civis Platform workflows:
https://support.civisanalytics.com/hc/en-us/articles/115004172983-Workflows-Basics

Mistral Workflow Language v2:
https://docs.openstack.org/mistral/latest/user/wf_lang_v2.html#workflows
"""

from __future__ import annotations

import inspect

from civis import APIClient
from civis.resources import API_SPEC_PATH


_CLIENT = APIClient(local_api_spec=API_SPEC_PATH, api_key="no-key-needed")


def _endpoint_method_params(endpoint: str, method: str) -> tuple[list[str], list[str]]:
    endpt = getattr(_CLIENT, endpoint)
    meth = getattr(endpt, method)
    method_params = inspect.signature(meth).parameters
    required, optional = [], []
    for name, param in method_params.items():
        if param.default == inspect.Parameter.empty:
            required.append(name)
        else:
            optional.append(name)
    return required, optional


def _if_then_create_script(action: str) -> dict:
    endpoint, script_type = action.removeprefix("civis.").split(".")
    if script_type == "container":
        script_type = "containers"
    required, optional = _endpoint_method_params(endpoint, f"post_{script_type}")
    if "name" in required:
        # Civis Platform allows the workflow task name to be the script name.
        required.remove("name")
        optional.append("name")
    return {
        "if": {"properties": {"action": {"const": action}}},
        "then": {
            "properties": {
                "input": {
                    "type": "object",
                    # Although we have type annotations for each key name,
                    # leave the value unspecified as {} to allow YAQL expressions.
                    "properties": {name: {} for name in required + optional},
                    "required": required,
                    "additionalProperties": False,
                },
            },
        },
    }


def _if_then_import() -> dict:
    required_post, optional_post = _endpoint_method_params("imports", "post")
    required_post_syncs, optional_post_syncs = _endpoint_method_params(
        "imports", "post_syncs"
    )
    if "name" in required_post:
        # Civis Platform allows the workflow task name to be the script name.
        required_post.remove("name")
        optional_post.append("name")
    if "id" in required_post_syncs:
        # The "id" will come from the job ID of the "post" call.
        required_post_syncs.remove("id")
    properties = {
        **{name: {} for name in required_post + optional_post},
        "syncs": {
            "type": "array",
            "items": {
                "type": "object",
                # Although we have type annotations for each key name,
                # leave the value unspecified as {} to allow YAQL expressions.
                "properties": {
                    name: {} for name in required_post_syncs + optional_post_syncs
                },
                "required": required_post_syncs,
                "additionalProperties": False,
            },
        },
    }
    return {
        "if": {"properties": {"action": {"const": "civis.import"}}},
        "then": {
            "properties": {
                "input": {
                    "type": "object",
                    "properties": properties,
                    "required": required_post,
                    "additionalProperties": False,
                },
            },
        },
    }


def _if_then_execute(action: str, id_name: str) -> dict:
    return {
        "if": {"properties": {"action": {"const": action}}},
        "then": {
            "properties": {
                "input": {
                    "type": "object",
                    # Although the ID should be an integer,
                    # leave it unspecified as {} to allow YAQL expressions.
                    "properties": {id_name: {}},
                    "required": [id_name],
                    "additionalProperties": False,
                },
            },
        },
    }


TASK_TRANSITION_SCHEMA = {
    "oneOf": [
        # A single task name.
        {"type": "string"},
        # A list of either (i) task names to transition to, or
        # (ii) task names that each have a YAQL guard expression, or
        # a mixture of (i) and (ii).
        {"type": "array", "items": {"oneOf": [{"type": "string"}, {"type": "object"}]}},
        # A single task name or a list of task names under the (optional?) key "next".
        {
            "type": "object",
            "properties": {
                "next": {
                    "oneOf": [
                        {"type": "string"},
                        {"type": "array", "items": {"type": "string"}},
                        {"type": "array", "items": {"type": "object"}},
                    ],
                },
            },
        },
    ],
}

TASK_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "maxLength": 255,
            "not": {"enum": ["noop", "fail", "succeed", "pause"]},
        },
        "description": {"type": "string"},
        "action": {
            "type": "string",
            "enum": [
                "civis.scripts.python3",
                "civis.scripts.r",
                "civis.scripts.sql",
                "civis.scripts.javascript",
                "civis.scripts.container",
                "civis.scripts.dbt",
                "civis.scripts.custom",
                "civis.enhancements.cass_ncoa",
                "civis.import",
                "civis.run_job",
                "civis.workflows.execute",
                "std.async_noop",
                "std.echo",
                "std.fail",
                "std.noop",
            ],
        },
        "input": {"type": "object"},
        "publish": {"type": "object"},
        "publish-on-error": {"type": "object"},
        "on-success": TASK_TRANSITION_SCHEMA,
        "on-error": TASK_TRANSITION_SCHEMA,
        "on-complete": TASK_TRANSITION_SCHEMA,
        "join": {
            "oneOf": [
                {"const": "all"},
                {"type": "integer", "minimum": 1},
            ],
        },
        "requires": {"type": "array"},
        "with-items": {
            "oneOf": [
                {"type": "string"},
                {"type": "array", "items": {"type": "string"}},
            ],
        },
        "keep-result": {"type": "boolean"},
        "target": {"type": "string"},
        "pause-before": {"type": "boolean"},
        "wait-before": {"type": "number", "minimum": 0},
        "wait-after": {"type": "number", "minimum": 0},
        "fail-on": {"type": "string"},
        "timeout": {"type": "number", "minimum": 0},
        "retry": {
            "oneOf": [
                {"type": "string"},
                {
                    "type": "object",
                    "properties": {
                        "count": {"type": "number", "minimum": 0},
                        "delay": {"type": "number", "minimum": 0},
                        "break-on": {"type": "string"},
                        "continue-on": {"type": "string"},
                    },
                },
            ],
        },
        "concurrency": {"type": "number", "minimum": 1},
        "safe-rerun": {"type": "boolean"},
    },
    "required": ["action"],
    "allOf": [
        # If "action" is one of the Civis-defined ones,
        # then the allowed properties under "input" closely mirror the relevant
        # API endpoint method.
        _if_then_create_script("civis.scripts.python3"),
        _if_then_create_script("civis.scripts.r"),
        _if_then_create_script("civis.scripts.sql"),
        _if_then_create_script("civis.scripts.javascript"),
        _if_then_create_script("civis.scripts.container"),
        _if_then_create_script("civis.scripts.dbt"),
        _if_then_create_script("civis.scripts.custom"),
        _if_then_create_script("civis.enhancements.cass_ncoa"),
        _if_then_execute("civis.run_job", "job_id"),
        _if_then_execute("civis.workflows.execute", "workflow_id"),
        _if_then_import(),
    ],
    "additionalProperties": False,
}


WORKFLOW_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {"version": {"const": "2.0"}},
    "patternProperties": {
        "^(?:(?!version).)*$": {
            "type": "object",
            "properties": {
                "type": {"type": "string"},
                "description": {"type": "string"},
                "input": {
                    "type": "array",
                    "items": {"oneOf": [{"type": "string"}, {"type": "object"}]},
                },
                "output": {},
                "output-on-error": {},
                "task-defaults": {
                    k: v for k, v in TASK_SCHEMA.items() if k != "required"
                },
                "tasks": {
                    "type": "object",
                    "patternProperties": {"^.*$": TASK_SCHEMA},
                    "minProperties": 1,
                },
            },
            "required": ["tasks"],
            "additionalProperties": True,  # Allow anchor definitions.
        },
    },
    "required": ["version"],
    "minProperties": 2,
    "maxProperties": 2,
}



from math import sqrt
import io
import pickle
from unittest import mock

import pytest
import requests
from joblib import delayed, Parallel, parallel_config, register_parallel_backend

import civis.parallel
from civis.base import CivisAPIError, CivisJobFailure
from civis.response import Response
from civis.futures import ContainerFuture
from civis.tests import create_client_mock, create_client_mock_for_container_tests


_MOCK_JOB_KWARGS = dict(
    from_template_id=None,
    id="42",
    required_resources={"cpu": 11},
    docker_image_name="image_name",
    docker_image_tag="tag",
    repo_http_uri="cabbage",
    repo_ref="servant",
    remote_host_credential_id=171,
    git_credential_id=213,
    cancel_timeout=23,
    time_zone="America/Chicago",
)


@pytest.fixture
def mock_job():
    return Response(
        dict(params=[{"name": "spam"}], arguments={"spam": "eggs"}, **_MOCK_JOB_KWARGS)
    )


@pytest.fixture
def mock_child_job():
    params = [
        {"name": "spam"},
        {"name": "CIVIS_PARENT_JOB_ID", "value": "123"},
        {"name": "CIVIS_PARENT_RUN_ID", "value": "456"},
    ]
    args = {"spam": "eggs"}
    return Response(dict(params=params, arguments=args, **_MOCK_JOB_KWARGS))


@pytest.mark.parametrize(
    "num_failures,max_submit_retries,submit_call_count,should_fail,from_template_id",
    [
        # Test that submission doesn't fail when there are no mock API errors.
        (0, 0, 3, False, None),
        (0, 5, 3, False, None),
        # Test that submission fails when there are API errors and too few retries.
        (2, 0, 1, True, None),
        (2, 1, 1, True, None),
        (2, 2, 2, True, None),
        # Test that submission doesn't fail when there are mock API errors and
        # sufficient retries.
        (2, 3, 5, False, None),
        (2, 4, 5, False, None),
        # Using a template.
        (0, 5, 3, False, 13),
        (1, 5, 4, False, 13),
        (1, 0, 1, True, 13),
    ],
)
@mock.patch("civis.parallel._robust_pickle_download")
@mock.patch("civis.io._files.requests")
@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
@mock.patch.object(civis.parallel, "_ContainerShellExecutor")
@mock.patch.object(civis.parallel, "CustomScriptExecutor")
def test_retries(
    mock_custom_exec_cls,
    mock_executor_cls,
    mock_sleep,
    mock_requests,
    mock_download,
    num_failures,
    max_submit_retries,
    submit_call_count,
    should_fail,
    from_template_id,
):

    mock_client = create_client_mock()
    mock_client.scripts.get_containers_runs.return_value.state = "succeeded"
    mock_client.files.post.return_value.id = 12345

    mock_requests_response = mock.Mock()
    mock_requests_response.ok = True
    mock_requests.post.return_value = mock_requests_response

    mock_download.side_effect = [[0.0], [1.0], [2.0]]

    # A function to raise fake API errors the first num_failures times it is
    # called.
    counter = {"n_failed": 0}

    def mock_submit(fn="", *args, **kwargs):
        if counter["n_failed"] < num_failures:
            counter["n_failed"] += 1
            raise CivisAPIError(mock.MagicMock())
        else:
            future = ContainerFuture(1, 2, client=mock_client)
            future._result = Response({"state": "succeeded"})
            return future

    mock_custom_exec_cls.return_value.submit.side_effect = mock_submit
    mock_executor_cls.return_value.submit.side_effect = mock_submit

    if from_template_id:
        factory = civis.parallel.make_backend_template_factory(
            from_template_id=from_template_id,
            max_submit_retries=max_submit_retries,
            client=mock_client,
        )
    else:
        factory = civis.parallel.make_backend_factory(
            max_submit_retries=max_submit_retries, client=mock_client
        )
    register_parallel_backend("civis", factory)
    with parallel_config("civis"):
        # NB: Since joblib v1.3.0,
        # n_jobs=1 would just run a for loop and wouldn't use any parallel backend.
        parallel = Parallel(n_jobs=2, pre_dispatch="n_jobs")
        if should_fail:
            with pytest.raises(civis.parallel.JobSubmissionError):
                parallel(delayed(sqrt)(i**2) for i in range(3))
            if from_template_id:
                assert (
                    mock_custom_exec_cls.return_value.submit.call_count
                    == submit_call_count
                )
                mock_custom_exec_cls.return_value.submit.assert_called_with(
                    JOBLIB_FUNC_FILE_ID=12345
                )
            else:
                assert (
                    mock_executor_cls.return_value.submit.call_count
                    == submit_call_count
                )
        else:
            result = parallel(delayed(sqrt)(i**2) for i in range(3))
            assert result == [0.0, 1.0, 2.0]
            if from_template_id:
                assert (
                    mock_custom_exec_cls.return_value.submit.call_count
                    == submit_call_count
                )
                mock_custom_exec_cls.return_value.submit.assert_called_with(
                    JOBLIB_FUNC_FILE_ID=12345
                )
            else:
                assert (
                    mock_executor_cls.return_value.submit.call_count
                    == submit_call_count
                )


@mock.patch.object(civis.parallel, "_CivisBackend")
def test_make_template(mock_backend):
    # Verify that the input setup command is recognized
    func = civis.parallel.make_backend_template_factory(1234)

    assert mock_backend.call_count == 0
    func()
    assert mock_backend.call_count == 1
    assert mock_backend.call_args_list[0][1].get("from_template_id") == 1234


@mock.patch.object(civis.parallel, "_CivisBackend")
def test_setup_cmd(mock_backend):
    # Verify that the input setup command is recognized
    func = civis.parallel.make_backend_factory(setup_cmd="sample")

    assert mock_backend.call_count == 0
    func()
    assert mock_backend.call_count == 1
    assert mock_backend.call_args_list[0][1].get("setup_cmd") == "sample"


@mock.patch.object(civis.parallel, "_CivisBackend")
def test_default_setup_cmd_no_repo(mock_backend):
    # Check that the setup command has the expected
    # default when the user does not input a GitHub repo
    func = civis.parallel.make_backend_factory()

    func()
    assert mock_backend.call_count == 1
    assert mock_backend.call_args_list[0][1].get("setup_cmd") == ":"


@mock.patch.object(civis.parallel, "_CivisBackend")
def test_default_setup_cmd_with_repo(mock_backend):
    # Check that the default setup command will attempt to install
    # a supplied GitHub repo.
    func = civis.parallel.make_backend_factory(repo_http_uri="potato")

    func()
    assert mock_backend.call_count == 1
    assert (
        mock_backend.call_args_list[0][1].get("setup_cmd")
        == "cd /app; pip install .; cd /"
    )


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_no_job_id_error(mock_make_factory, mock_job):
    # The `infer_backend_factory` should give a RuntimeError
    # if there's no CIVIS_JOB_ID in the environment.
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_job
    with mock.patch.dict("os.environ", {}, clear=True):
        with pytest.raises(RuntimeError):
            civis.parallel.infer_backend_factory(client=mock_client)


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer(mock_make_factory, mock_job):
    # Verify that `infer_backend_factory` passes through
    # the expected arguments to `make_backend_factory`.
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_job
    with mock.patch.dict(
        "os.environ", {"CIVIS_JOB_ID": "test_job", "CIVIS_RUN_ID": "test_run"}
    ):
        civis.parallel.infer_backend_factory(client=mock_client)

    expected = mock_job.json()
    del expected["from_template_id"]
    del expected["id"]
    mock_make_factory.assert_called_once_with(
        client=mock_client,
        setup_cmd=None,
        polling_interval=None,
        max_submit_retries=0,
        max_job_retries=0,
        hidden=True,
        remote_backend="sequential",
        **expected,
    )


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_new_params(mock_make_factory, mock_job):
    # Test overwriting existing job parameters with new parameters
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_job
    new_params = [{"name": "spam", "type": "fun"}, {"name": "foo", "type": "bar"}]
    with mock.patch.dict(
        "os.environ", {"CIVIS_JOB_ID": "test_job", "CIVIS_RUN_ID": "test_run"}
    ):
        civis.parallel.infer_backend_factory(client=mock_client, params=new_params)

    assert mock_make_factory.call_args[1]["params"] == new_params


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_extra_param(mock_make_factory, mock_job):
    # Test adding a new parameter and keeping
    # the existing parameter unchanged.
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_job
    new_params = [{"name": "foo", "type": "bar"}]
    with mock.patch.dict(
        "os.environ", {"CIVIS_JOB_ID": "test_job", "CIVIS_RUN_ID": "test_run"}
    ):
        civis.parallel.infer_backend_factory(client=mock_client, params=new_params)

    assert mock_make_factory.call_args[1]["params"] == [
        {"name": "spam"},
        {"name": "foo", "type": "bar"},
    ]


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_update_resources(mock_make_factory, mock_job):
    # Verify that users can modify requested resources for jobs.
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_job
    with mock.patch.dict(
        "os.environ", {"CIVIS_JOB_ID": "test_job", "CIVIS_RUN_ID": "test_run"}
    ):
        civis.parallel.infer_backend_factory(
            client=mock_client, required_resources={"cpu": -11}
        )

    assert mock_make_factory.call_args[1]["required_resources"] == {"cpu": -11}


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_update_args(mock_make_factory, mock_job):
    # Verify that users can modify the existing job's
    # arguments for sub-processes.
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_job
    with mock.patch.dict(
        "os.environ", {"CIVIS_JOB_ID": "test_job", "CIVIS_RUN_ID": "test_run"}
    ):
        civis.parallel.infer_backend_factory(
            client=mock_client, arguments={"foo": "bar"}
        )

    assert mock_make_factory.call_args[1]["arguments"] == {"spam": "eggs", "foo": "bar"}


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_from_custom_job(mock_make_factory, mock_job):
    # Test that `infer_backend_factory` can find needed
    # parameters if it's run inside a custom job created
    # from a template.
    mock_client = create_client_mock()
    mock_custom = Response(
        dict(
            from_template_id=999,
            id=42,
            required_resources=None,
            params=[{"name": "spam"}],
            arguments={"spam": "eggs"},
            docker_image_name="image_name",
            docker_image_tag="tag",
            repo_http_uri="cabbage",
            repo_ref="servant",
        )
    )
    mock_template = Response(dict(id=999, script_id=171))

    def _get_container(job_id):
        if int(job_id) == 42:
            return mock_custom
        elif int(job_id) == 171:
            return mock_job
        else:
            raise ValueError("Got job_id {}".format(job_id))

    mock_client.scripts.get_containers.side_effect = _get_container
    mock_client.templates.get_scripts.return_value = mock_template
    with mock.patch.dict(
        "os.environ", {"CIVIS_JOB_ID": "42", "CIVIS_RUN_ID": "test_run"}
    ):
        civis.parallel.infer_backend_factory(client=mock_client)

    # We should have called `get_containers` twice now -- once for
    # the container we're running in, and a second time for the
    # container which backs the template this job was created from.
    # The backing script has settings which aren't visible from
    # the container which was created from it.
    assert mock_client.scripts.get_containers.call_count == 2
    mock_client.templates.get_scripts.assert_called_once_with(999)
    expected_kwargs = {
        "required_resources": {"cpu": 11},
        "params": [{"name": "spam"}],
        "arguments": {"spam": "eggs"},
        "client": mock.ANY,
        "polling_interval": mock.ANY,
        "setup_cmd": None,
        "max_submit_retries": mock.ANY,
        "max_job_retries": mock.ANY,
        "hidden": True,
        "remote_backend": "sequential",
    }
    for key in civis.parallel.KEYS_TO_INFER:
        expected_kwargs[key] = mock_job[key]
    mock_make_factory.assert_called_once_with(**expected_kwargs)


@mock.patch.object(civis.parallel, "make_backend_factory")
def test_infer_in_child_job(mock_make_factory, mock_child_job):
    # Verify that infer_backend_factory doesn't include CIVIS_PARENT_JOB_ID and
    # CIVIS_PARENT_RUN_ID since those will be automatically added later.
    mock_client = create_client_mock()
    mock_client.scripts.get_containers.return_value = mock_child_job
    mock_env = {"CIVIS_JOB_ID": "test_job", "CIVIS_RUN_ID": "test_run"}
    with mock.patch.dict("os.environ", mock_env):
        civis.parallel.infer_backend_factory(client=mock_client)

    assert mock_make_factory.call_args[1]["params"] == [{"name": "spam"}]


def make_to_file_mock(result, max_n_err=0, exc=None):
    cnt = {"err": 0}

    def mock_civis_to_file(file_id, buf, client=None):
        if cnt["err"] < max_n_err:
            cnt["err"] += 1
            raise exc
        else:
            buf.write(pickle.dumps(result))

    return mock_civis_to_file


@mock.patch.object(civis.parallel, "civis")
def test_result_success(mock_civis):
    # Test that we can get a result back from a succeeded job.
    callback = mock.MagicMock()
    mock_civis.io.civis_to_file.side_effect = make_to_file_mock("spam")
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="success", run_outputs=mock.MagicMock()
    )
    fut = ContainerFuture(1, 2, client=mock_client)
    res = civis.parallel._CivisBackendResult(fut, callback)

    assert res.get() == "spam"
    assert callback.call_count == 1


@mock.patch.object(civis.parallel, "civis")
def test_result_callback_no_get(mock_civis):
    # Test that the completed callback happens even if we don't call `get`
    callback = mock.MagicMock()
    mock_civis.io.civis_to_file.side_effect = make_to_file_mock("spam")
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="success", run_outputs=mock.MagicMock()
    )
    fut = ContainerFuture(1, 2, client=mock_client)
    civis.parallel._CivisBackendResult(fut, callback)
    assert callback.call_count == 1


@mock.patch.object(civis.parallel, "civis")
@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_result_exception(m_sleep, mock_civis):
    # An error in the job should be raised by the result
    callback = mock.MagicMock()
    exc = ZeroDivisionError()
    mock_civis.io.civis_to_file.side_effect = make_to_file_mock(exc)
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="failed", run_outputs=mock.MagicMock()
    )
    fut = ContainerFuture(1, 2, client=mock_client)
    res = civis.parallel._CivisBackendResult(fut, callback)

    with pytest.raises(ZeroDivisionError):
        res.get()
    assert callback.call_count == 0


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_result_exception_no_result(m_sleep):
    # If the job errored but didn't write an output, we should get
    # a CivisJobFailure back.
    callback = mock.MagicMock()

    mock_client = create_client_mock_for_container_tests(
        1, 2, state="failed", run_outputs=[]
    )
    fut = ContainerFuture(1, 2, client=mock_client)
    res = civis.parallel._CivisBackendResult(fut, callback)
    fut._set_api_exception(CivisJobFailure(Response({"state": "failed"})))

    with pytest.raises(CivisJobFailure) as exc:
        res.get()

    assert "Response({'state': 'failed'})" in str(exc.value)
    assert callback.call_count == 0


@mock.patch.object(civis.parallel, "civis")
def test_result_callback_exception(mock_civis):
    # An error in the result retrieval should be raised by .get
    callback = mock.MagicMock()
    exc = ZeroDivisionError()
    mock_civis.io.civis_to_file.side_effect = exc
    # We're simulating a job which succeeded but generated an
    # exception when we try to download the outputs.
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="succeeded", run_outputs=mock.MagicMock()
    )
    fut = ContainerFuture(1, 2, client=mock_client)

    res = civis.parallel._CivisBackendResult(fut, callback)

    with pytest.raises(ZeroDivisionError):
        res.get()
    assert callback.call_count == 0


@mock.patch.object(civis.parallel, "civis")
@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_result_eventual_success(m_sleep, mock_civis):
    # Test that we can get a result back from a succeeded job,
    # even if we need to retry a few times to succeed with the download.
    callback = mock.MagicMock()
    exc = requests.ConnectionError()
    se = make_to_file_mock("spam", max_n_err=2, exc=exc)
    mock_civis.io.civis_to_file.side_effect = se
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="success", run_outputs=mock.MagicMock()
    )
    fut = ContainerFuture(1, 2, client=mock_client)
    res = civis.parallel._CivisBackendResult(fut, callback)

    assert res.get() == "spam"
    assert callback.call_count == 1


@mock.patch.object(civis.parallel, "civis")
@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_result_eventual_failure(m_sleep, mock_civis):
    # We will retry a connection error up to 5 times. Make sure
    # that we will get an error if it persists forever.
    callback = mock.MagicMock()
    exc = requests.ConnectionError()
    se = make_to_file_mock("spam", max_n_err=10, exc=exc)
    mock_civis.io.civis_to_file.side_effect = se
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="success", run_outputs=mock.MagicMock()
    )
    fut = ContainerFuture(1, 2, client=mock_client)
    res = civis.parallel._CivisBackendResult(fut, callback)
    with pytest.raises(requests.ConnectionError):
        res.get()
    assert callback.call_count == 0


@mock.patch.object(civis.parallel, "civis")
def test_result_running_and_cancel_requested(mock_civis):
    # When scripts request cancellation, they remain in a running
    # state. Make sure these are treated as cancelled runs.
    response = Response({"is_cancel_requested": True, "state": "running"})
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="running", run_outputs=mock.MagicMock()
    )
    mock_client.scripts.post_cancel.return_value = response
    fut = ContainerFuture(1, 2, client=mock_client)
    callback = mock.MagicMock()
    # When a _CivisBackendResult created by the Civis joblib backend completes
    # successfully, a callback is executed. When cancelled, this callback
    # shouldn't  be run
    civis.parallel._CivisBackendResult(fut, callback)
    fut.cancel()

    assert callback.call_count == 0


@mock.patch.object(civis.parallel, "civis")
@mock.patch.object(civis.parallel, "register_parallel_backend")
def test_setup_remote_backend(mock_register, mock_civis):
    backend = civis.parallel._CivisBackend()
    backend_name = civis.parallel._setup_remote_backend(backend)
    assert backend_name == "civis"
    assert mock_register.call_count == 1


def test_civis_backend_from_existing():
    # Test to make sure that making a new backend from an existing one makes
    # an exact copy.
    backend = civis.parallel._CivisBackend(
        setup_cmd="blah",
        from_template_id=-1,
        max_submit_retries=10,
        client="ha",
        remote_backend="cool",
        hidden=False,
    )

    new_backend = civis.parallel._CivisBackend.from_existing(backend)

    assert new_backend.setup_cmd == "blah"
    assert new_backend.from_template_id == -1
    assert new_backend.max_submit_retries == 10
    assert new_backend.client == "ha"
    assert new_backend.remote_backend == "cool"
    assert new_backend.executor_kwargs == {"hidden": False}


@mock.patch.object(civis.parallel, "civis")
def test_civis_backend_pickles(mock_civis):
    # Test to make sure the backend will pickle.
    backend = civis.parallel._CivisBackend(
        setup_cmd="blah",
        from_template_id=-1,
        max_submit_retries=10,
        client="ha",
        remote_backend="cool",
        hidden=False,
    )

    with parallel_config(backend):
        Parallel(n_jobs=-1)([])

    buff = io.BytesIO()
    pickle.dump(backend, buff)
    buff.seek(0)
    new_backend = pickle.load(buff)

    with parallel_config(new_backend):
        Parallel(n_jobs=-1)([])



import logging
from uuid import uuid4

from civis import civis_logger


def _get_test_logger(*args, **kwargs):
    # Need to use a logger of a different name in each test function,
    # or else we'd hit this issue:
    # https://github.com/pytest-dev/pytest/issues/5577
    logger = civis_logger(name=str(uuid4()), *args, **kwargs)
    # Set `propagate` back to `True`,
    # or else all the logging/caplog tests would fail.
    logger.propagate = True
    return logger


def test_civis_logger_base_case(caplog, capsys):
    log = _get_test_logger()
    caplog.set_level(log.level)

    log.debug("debug level")
    log.info("this is info level")
    log.warning("warning!")
    log.error("error!!")

    actual_logs = [(rec.levelname, rec.message) for rec in caplog.records]
    expected_logs = [
        ("INFO", "this is info level"),
        ("WARNING", "warning!"),
        ("ERROR", "error!!"),
    ]
    assert actual_logs == expected_logs

    captured = capsys.readouterr()
    assert captured.out == "this is info level\n"
    assert captured.err == "warning!\nerror!!\n"


def test_civis_logger_set_to_debug_level(caplog, capsys):
    log = _get_test_logger(level=logging.DEBUG)
    caplog.set_level(log.level)

    log.debug("debug level")
    log.info("this is info level")
    log.warning("warning!")
    log.error("error!!")

    actual_logs = [(rec.levelname, rec.message) for rec in caplog.records]
    expected_logs = [
        ("DEBUG", "debug level"),
        ("INFO", "this is info level"),
        ("WARNING", "warning!"),
        ("ERROR", "error!!"),
    ]
    assert actual_logs == expected_logs

    captured = capsys.readouterr()
    assert captured.out == "debug level\nthis is info level\n"
    assert captured.err == "warning!\nerror!!\n"


def test_civis_logger_fmt_from_str(caplog, capsys):
    log = _get_test_logger(fmt="%(levelname)s:%(message)s")
    caplog.set_level(log.level)

    log.debug("debug level")
    log.info("this is info level")
    log.warning("warning!")
    log.error("error!!")

    actual_logs = [(rec.levelname, rec.message) for rec in caplog.records]
    expected_logs = [
        ("INFO", "this is info level"),
        ("WARNING", "warning!"),
        ("ERROR", "error!!"),
    ]
    assert actual_logs == expected_logs

    captured = capsys.readouterr()
    assert captured.out == "INFO:this is info level\n"
    assert captured.err == "WARNING:warning!\nERROR:error!!\n"


def test_civis_logger_fmt_from_formatter(caplog, capsys):
    fmt = logging.Formatter("%(levelname)s:%(message)s")
    log = _get_test_logger(fmt=fmt)
    caplog.set_level(log.level)

    log.debug("debug level")
    log.info("this is info level")
    log.warning("warning!")
    log.error("error!!")

    actual_logs = [(rec.levelname, rec.message) for rec in caplog.records]
    expected_logs = [
        ("INFO", "this is info level"),
        ("WARNING", "warning!"),
        ("ERROR", "error!!"),
    ]
    assert actual_logs == expected_logs

    captured = capsys.readouterr()
    assert captured.out == "INFO:this is info level\n"
    assert captured.err == "WARNING:warning!\nERROR:error!!\n"



from datetime import datetime, timedelta
from unittest.mock import Mock, patch

import pytest

import civis
from civis.response import Response
from civis.tests.mocks import create_client_mock
from civis.utils._jobs import (
    _LOGS_PER_QUERY,
    _compute_effective_max_log_id,
    _job_finished_past_timeout,
    job_logs,
    logging,
)


def create_mock_client_with_job():
    mock_client = create_client_mock()

    job_response = Response({"id": 1, "name": "test"})
    run_post_response = Response({"id": 1})
    run_get_response = Response(
        {
            "id": run_post_response.id,
            "state": "succeeded",
            "is_cancel_requested": False,
            "error": None,
            "custom_id": run_post_response.id,
        }
    )
    mock_client.scripts.post_custom.return_value = job_response
    mock_client.scripts.post_custom_runs.return_value = run_post_response
    mock_client.scripts.get_custom_runs.return_value = run_get_response
    return mock_client


@pytest.fixture
def mock_client():
    return create_client_mock()


@pytest.fixture
def mock_client_single_json_output():
    mock_client = create_mock_client_with_job()
    mock_output = [
        Response(
            {
                "object_type": "JSONValue",
                "object_id": 10,
                "name": "output",
                "value": {"a": 1},
            }
        )
    ]
    mock_client.scripts.list_custom_runs_outputs.return_value = mock_output
    return mock_client


@pytest.fixture
def mock_client_multiple_json_output():
    mock_client = create_mock_client_with_job()
    mock_output = [
        Response(
            {
                "object_type": "JSONValue",
                "object_id": 10,
                "name": "output",
                "value": {"a": 1},
            }
        ),
        Response(
            {
                "object_type": "JSONValue",
                "object_id": 11,
                "name": "output2",
                "value": {"b": 2},
            }
        ),
    ]
    mock_client.scripts.list_custom_runs_outputs.return_value = mock_output
    return mock_client


@pytest.fixture
def mock_client_no_json_output():
    mock_client = create_mock_client_with_job()
    mock_output = [
        Response(
            {
                "object_type": int,
                "object_id": 10,
                "name": "output",
                "value": 100,
            }
        )
    ]
    mock_client.scripts.list_custom_runs_outputs.return_value = mock_output
    return mock_client


def test_run_template_json_output_fileids_returned(mock_client_single_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id, arguments=args, client=mock_client_single_json_output
    )
    assert result == {"output": 10}


def test_run_template_no_json_output_fileids_returned(mock_client_no_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id, arguments=args, client=mock_client_no_json_output
    )
    assert result == {"output": 10}


def test_run_template_multiple_json_output_fileids_returned(
    mock_client_multiple_json_output,
):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id, arguments=args, client=mock_client_multiple_json_output
    )
    assert result == {"output": 10, "output2": 11}


def test_run_template_json_returned(mock_client_single_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id,
        arguments=args,
        JSONValue=True,
        client=mock_client_single_json_output,
    )
    assert result == {"a": 1}


def test_run_template_many_json_outputs(caplog, mock_client_multiple_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id,
        arguments=args,
        JSONValue=True,
        client=mock_client_multiple_json_output,
    )
    log_result = caplog.record_tuples
    assert log_result == [
        (
            "civis.utils._jobs",
            logging.WARNING,
            "More than 1 JSON output for template {} "
            "-- returning only the first one.".format(template_id),
        )
    ]
    assert result == {"a": 1}


def test_run_template_when_no_json_output(caplog, mock_client_no_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id,
        arguments=args,
        JSONValue=True,
        client=mock_client_no_json_output,
    )
    log_result = caplog.record_tuples
    assert log_result == [
        (
            "civis.utils._jobs",
            logging.WARNING,
            "No JSON output for template {}".format(template_id),
        )
    ]
    assert result is None


@pytest.fixture
def sample_logs():
    return [
        {"id": 1, "message": "First log", "created_at": "2023-01-01T00:00:00Z"},
        {"id": 2, "message": "Second log", "created_at": "2023-01-01T00:00:01Z"},
        {"id": 3, "message": "Third log", "created_at": "2023-01-01T00:00:02Z"},
    ]


@patch("civis.utils._jobs.time")
def test_job_logs_single_batch(mock_time, mock_client, sample_logs):
    """Test when all logs are retrieved in a single batch."""
    mock_response = Mock()
    mock_response.json.return_value = sample_logs
    mock_response.headers = {"civis-max-id": "3", "civis-cache-control": "store"}

    mock_client.jobs.list_runs_logs.return_value = mock_response
    mock_time.time.return_value = datetime.fromisoformat("2025-01-01").timestamp()

    logs = list(job_logs(job_id=123, run_id=456, client=mock_client))

    assert len(logs) == 3
    assert logs == sample_logs
    mock_client.jobs.list_runs_logs.assert_called_once_with(
        123, 456, last_id=0, limit=_LOGS_PER_QUERY
    )


@patch("civis.utils._jobs.time")
def test_job_logs_multiple_batches(mock_time, mock_client):
    """Test when logs are retrieved in multiple batches."""
    first_batch = [
        {"id": 1, "message": "Log 1", "created_at": "2023-01-01T00:00:00Z"},
        {"id": 2, "message": "Log 2", "created_at": "2023-01-01T00:00:01Z"},
    ]
    second_batch = [{"id": 3, "message": "Log 3", "created_at": "2023-01-01T00:00:02Z"}]

    mock_response1 = Mock()
    mock_response1.json.return_value = first_batch
    mock_response1.headers = {"civis-max-id": "2", "civis-cache-control": "no-store"}

    mock_response2 = Mock()
    mock_response2.json.return_value = second_batch
    mock_response2.headers = {"civis-max-id": "3", "civis-cache-control": "store"}

    mock_client.jobs.list_runs_logs.side_effect = [mock_response1, mock_response2]
    mock_time.time.return_value = datetime.fromisoformat("2025-01-01").timestamp()

    logs = list(job_logs(job_id=123, run_id=456, client=mock_client))

    assert len(logs) == 3
    assert logs == first_batch + second_batch
    assert mock_client.jobs.list_runs_logs.call_count == 2
    mock_time.sleep.assert_called_once()


@patch("civis.utils._jobs.time")
def test_job_logs_no_logs_initially(mock_time, mock_client, sample_logs):
    """Test behavior when no logs are available."""

    mock_response1 = Mock()
    mock_response1.json.return_value = []
    mock_response1.headers = {"civis-cache-control": "store"}

    mock_response2 = Mock()
    mock_response2.json.return_value = sample_logs
    mock_response2.headers = {"civis-cache-control": "store", "civis-max-id": "3"}

    mock_client.jobs.list_runs_logs.side_effect = [mock_response1, mock_response2]
    mock_time.time.return_value = datetime.fromisoformat("2025-01-01").timestamp()

    logs = list(job_logs(job_id=123, run_id=456, client=mock_client))

    assert len(logs) == len(sample_logs)
    assert mock_client.jobs.list_runs_logs.call_count == 2
    mock_time.sleep.assert_called_once()


def test_job_logs_sorted_order(mock_client):
    """Test that logs are properly sorted by createdAt and id.

    Note: logs won't be sorted if they are out of order across different API calls.
    The sorting only works within a single API call.
    """
    unsorted_logs = [
        {"id": 2, "message": "Log 2", "created_at": "2023-01-01T00:00:01Z"},
        {"id": 1, "message": "Log 1", "created_at": "2023-01-01T00:00:01Z"},
        {"id": 3, "message": "Log 3", "created_at": "2023-01-01T00:00:00Z"},
    ]

    expected_order = [
        {"id": 3, "message": "Log 3", "created_at": "2023-01-01T00:00:00Z"},
        {"id": 1, "message": "Log 1", "created_at": "2023-01-01T00:00:01Z"},
        {"id": 2, "message": "Log 2", "created_at": "2023-01-01T00:00:01Z"},
    ]

    mock_response = Mock()
    mock_response.json.return_value = unsorted_logs
    mock_response.headers = {"civis-max-id": "3", "civis-cache-control": "store"}

    mock_client.jobs.list_runs_logs.return_value = mock_response

    logs = list(job_logs(job_id=123, run_id=456, client=mock_client))

    assert logs == expected_order


@patch("civis.utils._jobs.time")
def test_job_logs_no_duplicate_logs(mock_time, mock_client):
    """Test that duplicate log messages won't be yielded."""
    expected_logs = [
        {"id": 1, "message": "Log 1", "created_at": "2023-01-01T00:00:00Z"},
        {"id": 2, "message": "Log 2", "created_at": "2023-01-01T00:00:01Z"},
    ]

    mock_response1 = Mock()
    mock_response1.json.return_value = expected_logs
    mock_response1.headers = {"civis-max-id": "2", "civis-cache-control": "no-store"}

    mock_response2 = Mock()
    mock_response2.json.return_value = expected_logs
    mock_response2.headers = {"civis-max-id": "2", "civis-cache-control": "store"}

    mock_time.time.return_value = datetime.fromisoformat("2025-01-01").timestamp()
    mock_client.jobs.list_runs_logs.side_effect = (mock_response1, mock_response2)

    actual_logs = list(job_logs(job_id=123, run_id=456, client=mock_client))

    assert actual_logs == expected_logs


def test_compute_effective_max_log_id_empty_logs():
    assert _compute_effective_max_log_id([]) == 0


@patch("civis.utils._jobs.time")
def test_compute_effective_max_log_id_all_logs_before_cutoff(mock_time):
    dt_now = datetime.now()
    # The log messages were from more seconds ago than the cutoff.
    logs = [
        {"id": 1, "created_at": (dt_now - timedelta(seconds=400)).isoformat()},
        {"id": 2, "created_at": (dt_now - timedelta(seconds=350)).isoformat()},
    ]
    mock_time.time.side_effect = lambda: dt_now.timestamp()
    assert _compute_effective_max_log_id(logs) == 2


@patch("civis.utils._jobs.time")
def test_compute_effective_max_log_id_logs_within_cutoff(mock_time):
    dt_now = datetime.now()
    # Both log messages were from fewer seconds ago than the cutoff,
    # so they'll both be retrieved again to avoid skipping any.
    logs = [
        {"id": 1, "created_at": (dt_now - timedelta(seconds=200)).isoformat()},
        {"id": 2, "created_at": (dt_now - timedelta(seconds=150)).isoformat()},
    ]
    mock_time.time.side_effect = lambda: dt_now.timestamp()
    assert _compute_effective_max_log_id(logs) == 0


@patch("civis.utils._jobs.time")
def test_compute_effective_max_log_id_logs_exceed_refetch_count(mock_time):
    dt_now = datetime.now()
    logs = [
        {"id": i, "created_at": (dt_now - timedelta(seconds=i)).isoformat()}
        for i in range(110)
    ]
    mock_time.time.side_effect = lambda: dt_now.timestamp()
    assert _compute_effective_max_log_id(logs) == 10


def test_job_finished_past_timeout_no_timeout(mock_client):
    assert not _job_finished_past_timeout(123, 456, None, mock_client)


def test_job_finished_past_timeout_not_finished(mock_client):
    mock_client.jobs.get_runs.return_value.json.return_value = {"finished_at": None}
    assert not _job_finished_past_timeout(123, 456, 10, mock_client)


def test_job_finished_past_timeout_finished_recently(mock_client):
    finished_at = (datetime.now() - timedelta(seconds=5)).isoformat()
    mock_client.jobs.get_runs.return_value.json.return_value = {
        "finished_at": finished_at
    }
    assert not _job_finished_past_timeout(123, 456, 10, mock_client)


def test_job_finished_past_timeout_finished_long_ago(mock_client):
    finished_at = (datetime.now() - timedelta(seconds=20)).isoformat()
    mock_client.jobs.get_runs.return_value.json.return_value = {
        "finished_at": finished_at
    }
    assert _job_finished_past_timeout(123, 456, 10, mock_client)



from civis.io._databases import query_civis, transfer_table
from civis.io._files import (
    civis_to_file,
    dataframe_to_file,
    file_id_from_run_output,
    file_to_civis,
    file_to_dataframe,
    file_to_json,
    json_to_file,
)
from civis.io._tables import (
    civis_file_to_table,
    civis_to_csv,
    civis_to_multifile_csv,
    csv_to_civis,
    dataframe_to_civis,
    export_to_civis_file,
    read_civis,
    read_civis_sql,
    split_schema_tablename,
)


__all__ = [
    # from _databases.py
    "query_civis",
    "transfer_table",
    # From _files.py
    "civis_to_file",
    "dataframe_to_file",
    "file_id_from_run_output",
    "file_to_civis",
    "file_to_dataframe",
    "file_to_json",
    "json_to_file",
    # From _tables.py
    "civis_file_to_table",
    "civis_to_csv",
    "civis_to_multifile_csv",
    "csv_to_civis",
    "dataframe_to_civis",
    "export_to_civis_file",
    "read_civis",
    "read_civis_sql",
    "split_schema_tablename",
]



import io
import pickle
import pprint
from string import ascii_lowercase
from unittest import mock

import pytest
import requests

from civis.response import (
    CivisClientError,
    PaginatedResponse,
    _response_to_json,
    convert_response_data_type,
    Response,
    CivisImmutableResponseError,
    find,
)
from civis._camel_to_snake import camel_to_snake


def _create_mock_response(data, headers):
    mock_response = mock.MagicMock(spec=requests.Response)
    mock_response.json.return_value = data
    mock_response.headers = headers
    mock_response.status_code = 200
    return mock_response


def _create_empty_response(code, headers):
    mock_response = mock.MagicMock(spec=requests.Response)
    mock_response.status_code = code
    mock_response.content = b""
    mock_response.headers = headers
    return mock_response


def _make_paginated_response(path, params):
    results = [
        [
            {"id": 1, "name": "job_1"},
            {"id": 2, "name": "job_2"},
            {"id": 3, "name": "job_3"},
        ],
        [
            {"id": 4, "name": "job_4"},
            {"id": 5, "name": "job_5"},
        ],
        [],
    ]
    mock_endpoint = mock.MagicMock()
    mock_endpoint._make_request.side_effect = [
        _create_mock_response(result, {}) for result in results
    ]
    mock_endpoint._return_type = "snake"

    paginator = PaginatedResponse(path, params, mock_endpoint)

    return paginator, mock_endpoint


def test_pagination():
    path = "/objects"
    params = {"param": "value"}
    paginator, mock_endpoint = _make_paginated_response(path, params)

    # No API calls made yet.
    mock_endpoint._make_request.assert_not_called()

    all_data = []
    for indx, obj in enumerate(paginator):
        assert obj["id"] == indx + 1
        all_data.append(obj)

        # Test lazy evaluation. Should make only make one call up until the
        # first item of the second page is needed.
        if indx < 3:
            mock_endpoint._make_request.assert_called_once_with(
                "GET", path, dict(params, **{"page_num": 1})
            )
        else:
            mock_endpoint._make_request.assert_called_with(
                "GET", path, dict(params, **{"page_num": 2})
            )

    # One extra call is made. Pagination is stopped since the response is
    # empty.
    assert mock_endpoint._make_request.call_count == 3
    assert len(all_data) == 5


def test_iterator_interface():
    # Make sure that the PaginatedResponse implements `next` as expected
    paginator, _ = _make_paginated_response("/objects", {"param": "value"})

    assert next(paginator)["id"] == 1
    assert next(paginator)["id"] == 2
    assert next(paginator)["id"] == 3
    assert next(paginator)["id"] == 4
    assert next(paginator)["id"] == 5
    with pytest.raises(StopIteration):
        next(paginator)


def test_response_to_json_no_error():
    raw_response = _create_mock_response({"key": "value"}, None)
    assert _response_to_json(raw_response) == {"key": "value"}


def test_response_to_no_content_snake():
    # Test empty response handling for codes where we're likely to see them.
    for code in [202, 204, 205]:
        raw_response = _create_empty_response(code, {"header1": "val1"})
        data = convert_response_data_type(raw_response, return_type="snake")

        assert isinstance(data, Response)
        assert data.json_data is None
        assert data.headers == {"header1": "val1"}


def test_response_to_json_parsing_error():
    raw_response = mock.MagicMock()
    raw_response.json.side_effect = ValueError("Invalid json")
    with pytest.raises(CivisClientError) as excinfo:
        _response_to_json(raw_response)
    assert "Unable to parse JSON from response" in str(excinfo.value)


def test_convert_data_type_raw_unparsed():
    response = _create_mock_response({}, {"header1": "val1"})
    data = convert_response_data_type(response, return_type="raw")

    assert isinstance(data, requests.Response)
    assert data.headers == {"header1": "val1"}


def test_convert_data_type_raw_parsed():
    response = {"foo": "bar"}
    data = convert_response_data_type(response, return_type="raw")

    assert isinstance(data, dict)
    assert data == {"foo": "bar"}


def test_convert_data_type_civis():
    response = _create_mock_response({"foo": "bar"}, {"header": "val"})
    data = convert_response_data_type(response, return_type="snake")

    assert isinstance(data, Response)
    assert data["foo"] == "bar"
    assert data.headers == {"header": "val"}


def test_convert_data_type_civis_list():
    response = _create_mock_response(
        [{"foo": "bar"}, {"fizz": "buzz"}], {"header": "val"}
    )
    data = convert_response_data_type(response, return_type="snake")

    assert isinstance(data, list)
    assert len(data) == 2
    assert isinstance(data[0], Response)
    assert data[0]["foo"] == "bar"
    assert data[0].headers == {"header": "val"}


def test_parse_column_names():
    """Check that responses that include 'update' as a key are parsed right."""
    resp_dict = {
        "columns": [
            {
                "valueDistributionPercent": {"update": 50.0, "foo": 50.0},
                "valueDistribution": {"update": 1, "foo": 1},
            }
        ]
    }
    resp = Response(resp_dict)
    assert resp.columns[0].value_distribution_percent["update"] == 50.0


@pytest.mark.parametrize(
    "headers, expected_calls_remaining, expected_rate_limit",
    [
        (None, None, None),
        ({}, None, None),
        ({"X-RateLimit-Remaining": "1", "X-RateLimit-Limit": "100"}, 1, 100),
    ],
)
def test_rate_limit(headers, expected_calls_remaining, expected_rate_limit):
    response = Response({"foo": "bar"}, headers=headers)
    assert response.headers == headers
    assert response.calls_remaining == expected_calls_remaining
    assert response.rate_limit == expected_rate_limit


def test_response_is_immutable():
    """Test that the Response object is immutable.

    Resolves https://github.com/civisanalytics/civis-python/issues/228
    """
    # JSON data from the Civis API is in camelCase.
    json_data = {"fooBar": {"barBaz": "whatever"}}
    response = Response(json_data)
    assert response.json_data == json_data

    with pytest.raises(CivisImmutableResponseError):
        response["foo_bar"] = "something else"
    with pytest.raises(CivisImmutableResponseError):
        response.foo_bar = "something else"
    with pytest.raises(CivisImmutableResponseError):
        response["foo_bar"]["bar_baz"] = "something else"
    with pytest.raises(CivisImmutableResponseError):
        response.foo_bar.bar_baz = "something else"


def test_response_cross_compatibility():
    """Test cross compatibility: snake vs camel case, getitem vs getattr.

    Resolves https://github.com/civisanalytics/civis-python/issues/317
    """
    msg = "Life is a long journey. Enjoy it!"
    # JSON data from the Civis API is in camelCase.
    json_data = {"fooBar": {"barBaz": msg}}
    response = Response(json_data)
    assert response.json_data == json_data

    #   16 combinations altogether
    # = 2 ** 4
    # = {2 levels deep under the response} ** {snake/camel * getitem/getattr}
    assert (
        msg
        == response.foo_bar.bar_baz
        == response.foo_bar["bar_baz"]
        == response.foo_bar.barBaz
        == response.foo_bar["barBaz"]
        == response["foo_bar"].bar_baz
        == response["foo_bar"]["bar_baz"]
        == response["foo_bar"].barBaz
        == response["foo_bar"]["barBaz"]
        == response.fooBar.bar_baz
        == response.fooBar["bar_baz"]
        == response.fooBar.barBaz
        == response.fooBar["barBaz"]
        == response["fooBar"].bar_baz
        == response["fooBar"]["bar_baz"]
        == response["fooBar"].barBaz
        == response["fooBar"]["barBaz"]
    )


@pytest.mark.parametrize("key", ["arguments", "environmentVariables"])
def test_response_keys_preserve_case(key):
    json_data = {key: {"FOO": 123, "FOO_BAR": 456}}
    response = Response(json_data)
    resp = getattr(response, camel_to_snake(key))
    assert resp.FOO == resp["FOO"] == 123
    assert resp.FOO_BAR == resp["FOO_BAR"] == 456
    with pytest.raises(AttributeError):
        resp.foo
    with pytest.raises(KeyError):
        resp["foo"]
    with pytest.raises(AttributeError):
        resp.foo_bar
    with pytest.raises(KeyError):
        resp["foo_bar"]


@pytest.mark.parametrize(
    "source, as_snake_case",
    [
        ({"foo": {"barBar": 1}}, {"foo": {"bar_bar": 1}}),
        (
            {"fooBar": 1, "arguments": {"FOO": 2, "FOO_BAR": 3}},
            {"foo_bar": 1, "arguments": {"FOO": 2, "FOO_BAR": 3}},
        ),
        (
            {"fooBar": [{"name": "a", "type": "b"}, {"name": "c", "type": "d"}]},
            {"foo_bar": [{"name": "a", "type": "b"}, {"name": "c", "type": "d"}]},
        ),
        ({"fooBar": ["a", "b", "c"]}, {"foo_bar": ["a", "b", "c"]}),
    ],
)
def test_json(source, as_snake_case):
    response = Response(source)
    assert response.json() == as_snake_case
    assert response.json(snake_case=False) == source


def test_json_no_data():
    response = Response(None)
    assert response.json() == {}
    assert response.json(snake_case=False) == {}


@pytest.mark.parametrize("snake_case", [True, False])
def test_json_preserve_original_dict(snake_case):
    # User may want to modify the dict from response.json().
    # Make sure the dict from response.json() is not the same object as
    # the original dict passed to Response.
    json_data = {"foo": 123, "bar": 456}
    id_original = id(json_data)
    response = Response(json_data)
    id_in_response = id(response.json(snake_case=snake_case))
    assert id_original != id_in_response


@pytest.mark.parametrize(
    "json_data, expected_length",
    [
        ({}, 0),
        ({"foo": 123}, 1),
        ({"foo": 123, "bar": 456}, 2),
    ],
)
def test_len(json_data, expected_length):
    response = Response(json_data)
    assert response.json_data == json_data
    assert len(response) == expected_length


@pytest.mark.parametrize(
    "json_data, expected_repr",
    [
        (None, "Response({})"),
        ({}, "Response({})"),
        ({"foo": 123}, "Response({'foo': 123})"),
        ({"foo": {"barBaz": 456}}, "Response({'foo': Response({'bar_baz': 456})})"),
        # repr() call doesn't wrap long lines.
        (
            {
                "foo": {ascii_lowercase[i]: i for i in range(3)},
                "fooBar": {ascii_lowercase[i]: i for i in range(15)},
            },
            "Response({'foo': Response({'a': 0, 'b': 1, 'c': 2}), 'foo_bar': Response({'a': 0, 'b': 1, 'c': 2, 'd': 3, 'e': 4, 'f': 5, 'g': 6, 'h': 7, 'i': 8, 'j': 9, 'k': 10, 'l': 11, 'm': 12, 'n': 13, 'o': 14})})",  # noqa: E501
        ),
    ],
)
def test_repr(json_data, expected_repr):
    response = Response(json_data)
    assert response.json_data == json_data
    assert repr(response) == expected_repr


def test_repr_with_empty_string_key():
    response = Response({"": 123})
    assert repr(response) == "Response({'': 123})"


@pytest.mark.parametrize(
    "json_data, expected",
    [
        # A "short" response's pprint looks the same as its repr.
        ({"foo": {"barBaz": 456}}, "Response({'foo': Response({'bar_baz': 456})})"),
        # A "long" response's pprint triggers line wrapping in pretty-printing.
        (
            {ascii_lowercase[i]: i for i in range(15)},
            "Response({'a': 0,\n"
            "          'b': 1,\n"
            "          'c': 2,\n"
            "          'd': 3,\n"
            "          'e': 4,\n"
            "          'f': 5,\n"
            "          'g': 6,\n"
            "          'h': 7,\n"
            "          'i': 8,\n"
            "          'j': 9,\n"
            "          'k': 10,\n"
            "          'l': 11,\n"
            "          'm': 12,\n"
            "          'n': 13,\n"
            "          'o': 14})",
        ),
        (
            {
                "foo": {ascii_lowercase[i]: i for i in range(3)},
                "fooBar": {ascii_lowercase[i]: i for i in range(15)},
            },
            "Response({'foo': Response({'a': 0, 'b': 1, 'c': 2}),\n"
            "          'foo_bar': Response({'a': 0,\n"
            "                               'b': 1,\n"
            "                               'c': 2,\n"
            "                               'd': 3,\n"
            "                               'e': 4,\n"
            "                               'f': 5,\n"
            "                               'g': 6,\n"
            "                               'h': 7,\n"
            "                               'i': 8,\n"
            "                               'j': 9,\n"
            "                               'k': 10,\n"
            "                               'l': 11,\n"
            "                               'm': 12,\n"
            "                               'n': 13,\n"
            "                               'o': 14})})",
        ),
    ],
)
def test_pprint(json_data, expected):
    response = Response(json_data)
    assert pprint.pformat(response) == expected


def test_jsonvalue_as_run_output():
    json_data = {"objectType": "JSONValue", "value": {"foo": 456}}
    response = Response(json_data)
    value = response.value
    assert isinstance(value, dict)
    assert value == {"foo": 456}


def test_jsonvalue_as_response():
    # json_data comes from a client.json_values.{get,post,patch} call.
    json_data = {"id": 12345, "name": "JSON Value 12345", "value": {"a": 123}}
    response = Response(json_data, from_json_values=True)
    value = response.value
    assert isinstance(value, dict)
    assert value == {"a": 123}


def test_get():
    # JSON data from the Civis API is in camelCase.
    json_data = {"foo": 123, "bar": {"bazQux": 456}}
    response = Response(json_data)
    assert response.json_data == json_data
    assert response.get("foo") == 123
    assert response.bar.get("baz_qux") == response.bar.get("bazQux") == 456
    assert response.get("doesn't exist") is None
    assert response.get("doesn't exist", "fallback value") == "fallback value"
    assert response.bar.get("doesn't exist") is None


def test_items():
    # JSON data from the Civis API is in camelCase.
    json_data = {"foo": 123, "bar": {"bazQux": 456}}
    response = Response(json_data)
    assert response.json_data == json_data
    for k, v in response.items():
        assert (k, v) in (("foo", 123), ("bar", Response({"bazQux": 456})))


def test_eq():
    # JSON data from the Civis API is in camelCase.
    json_data = {"foo": 123, "bar": {"bazQux": 456}}
    response = Response(json_data)

    response2 = Response(json_data)
    assert response == response2
    assert response == {"foo": 123, "bar": {camel_to_snake("bazQux"): 456}}

    for not_response_or_dict in (789, "str", ["list"], ("tuple",), {"set"}, None):
        assert response != not_response_or_dict


def test_response_is_pickleable():
    # JSON data from the Civis API is in camelCase.
    json_data = {"foo": 123, "bar": 456}
    response = Response(json_data)
    pickled = pickle.dumps(response)
    unpickled = pickle.load(io.BytesIO(pickled))
    assert response == unpickled


def test_find_filter_with_kwargs():
    r1 = Response({"foo": 0, "bar": "a", "baz": True})
    r2 = Response({"foo": 1, "bar": "b", "baz": True})
    r3 = Response({"foo": 2, "bar": "b", "baz": False})

    assert find([r1, r2, r3], wrong_attr="whatever") == []

    assert find([r1, r2, r3], foo=0) == [r1]
    assert find([r1, r2, r3], foo=1) == [r2]
    assert find([r1, r2, r3], foo=1, bar="b") == [r2]

    assert find([r1, r2, r3], bar="b") == [r2, r3]
    assert find([r1, r2, r3], bar="b", foo=1) == [r2]

    assert find([r1, r2, r3], foo=True) == []
    assert find([r1, r2, r3], foo=False) == []
    assert find([r1, r2, r3], bar=True) == []
    assert find([r1, r2, r3], bar=False) == []
    assert find([r1, r2, r3], baz=True) == [r1, r2]
    assert find([r1, r2, r3], baz=False) == [r3]

    assert find([r1, r2, r3], foo=int) == [r2, r3]



import logging
import os
import sys


class _LogFilter(logging.Filter):
    def __init__(self, mode, level):
        super().__init__()
        if mode not in ("at_or_below", "above"):
            raise ValueError(f"mode must be one of {{at_or_below, above}}): {mode}")
        self.mode = mode
        self.level = level

    def filter(self, record):
        if self.mode == "at_or_below":
            return record.levelno <= self.level
        else:
            return record.levelno > self.level


def civis_logger(name=None, level=None, fmt="%(message)s"):
    """Return a logger for Civis Platform jobs.

    The logs of Civis Platform jobs format stdout in black and stderr in red.
    This logger sends INFO-level (or below) logging to stdout (black),
    and other levels' logging (WARNING, etc.) to stderr (red).

    Parameters
    ----------
    name : str, optional
        Logger name, to be passed into :func:`logging.getLogger`.
        A common practice is to pass in the module name as ``__name__``.
    level : int or str, optional
        Level from which logging is done,
        see https://docs.python.org/3/library/logging.html#logging-levels.
        If ``None`` or not provided, the level specified by the environment
        variable ``CIVIS_LOG_LEVEL`` is used
        (e.g., ``export CIVIS_LOG_LEVEL=DEBUG``).
        If this environment variable is also not given,
        the logging level defaults to ``logging.INFO``.
    fmt : str or logging.Formatter, optional
        Logging format. The default is ``"%(message)s"``.
        For the attributes that can be formatted, see:
        https://docs.python.org/3/library/logging.html#logrecord-objects
        Alternatively, you may pass in a :class:`logging.Formatter` instance
        for more custom formatting.

    Returns
    -------
    :class:`logging.Logger`
    """
    logger = logging.getLogger(name)

    if level is None:
        logger.setLevel(os.getenv("CIVIS_LOG_LEVEL") or logging.INFO)
    else:
        logger.setLevel(level)

    # When running on Civis Platform (as opposed to unit tests in CI),
    # we don't want to propagate log records to the root logger
    # in order to avoid duplicate logs.
    logger.propagate = False

    if isinstance(fmt, logging.Formatter):
        platform_fmt = fmt
    else:
        platform_fmt = logging.Formatter(fmt)

    at_or_below_info_hdlr = logging.StreamHandler(sys.stdout)
    at_or_below_info_hdlr.addFilter(_LogFilter("at_or_below", logging.INFO))
    at_or_below_info_hdlr.setFormatter(platform_fmt)
    logger.addHandler(at_or_below_info_hdlr)

    above_info_hdlr = logging.StreamHandler(sys.stderr)
    above_info_hdlr.addFilter(_LogFilter("above", logging.INFO))
    above_info_hdlr.setFormatter(platform_fmt)
    logger.addHandler(above_info_hdlr)

    return logger



from collections import defaultdict
from inspect import signature, isfunction

import civis
from civis.resources import generate_classes_maybe_cached


def download_latest_api_spec(path):
    client = civis.APIClient()
    try:
        job = client.scripts.post_custom(from_template_id=13448)
    except civis.base.CivisAPIError as e:
        if e.status_error == 404:
            raise EnvironmentError(
                "This script can only be run by a Civis employee with their "
                "regular Civis Platform account's API key."
            )
        else:
            raise
    fut = civis.utils.run_job(job.id, client=client, polling_interval=5)
    fut.result()
    print(f"custom script {fut.job_id} run {fut.run_id} has succeeded")
    outputs = client.scripts.list_custom_runs_outputs(fut.job_id, fut.run_id)
    file_id = civis.find_one(outputs, name="civis_api_spec.json").object_id
    with open(path, "wb") as f:
        civis.io.civis_to_file(file_id, f, client=client)


def _get_methods(cls) -> dict:
    return {
        method_name: method
        for method_name, method in vars(cls).items()
        if not method_name.startswith("_") and isfunction(method)
    }


def _compare(reference: dict, compared: dict) -> tuple[dict, dict]:
    new = {
        "endpoints": set(),
        "methods": defaultdict(set),
    }
    changed = {
        "method parameters": defaultdict(set),
        "method docstrings": defaultdict(set),
    }
    for endpoint_name in set(compared.keys()) - set(reference.keys()):
        new["endpoints"].add(endpoint_name)
    for endpoint_name in set(compared.keys()) & set(reference.keys()):
        methods_compared = _get_methods(compared[endpoint_name])
        methods_reference = _get_methods(reference[endpoint_name])
        if meth_names := (set(methods_compared.keys()) - set(methods_reference.keys())):
            for meth_name in meth_names:
                new["methods"][endpoint_name].add(meth_name)
        for meth_name in set(methods_compared.keys()) & set(methods_reference.keys()):
            method_compared = methods_compared[meth_name]
            method_reference = methods_reference[meth_name]
            if (
                signature(method_compared).parameters
                != signature(method_reference).parameters
            ):
                changed["method parameters"][endpoint_name].add(meth_name)
            if method_compared.__doc__ != method_reference.__doc__:
                changed["method docstrings"][endpoint_name].add(meth_name)
    # Convert defaultdicts to regular dicts for nicer pprinting.
    new["methods"] = dict(new["methods"])
    changed["method parameters"] = dict(changed["method parameters"])
    changed["method docstrings"] = dict(changed["method docstrings"])
    return new, changed


def compare_api_specs(path_current: str, path_upstream: str) -> tuple[dict, dict, dict]:
    """Compare two Civis API specs for whether there's a difference.

    Parameters
    ----------
    path_current : str
        Path of the current Civis API spec versioned in the civis-python codebase.
    path_upstream : str
        Path of the latest Civis API spec fetched from upstream.

    Returns
    -------
    tuple[dict, dict, dict]
        Dicts of added, removed, and changed endpoints and methods.
    """
    endpoints_current = generate_classes_maybe_cached(
        path_current, api_key="no_key_needed", api_version="1.0"
    )
    endpoints_upstream = generate_classes_maybe_cached(
        path_upstream, api_key="no_key_needed", api_version="1.0"
    )
    added, changed = _compare(endpoints_current, endpoints_upstream)
    removed, _ = _compare(endpoints_upstream, endpoints_current)
    return added, removed, changed



from ._resources import (
    generate_classes,
    get_api_spec,
    generate_classes_maybe_cached,
    CACHED_SPEC_PATH,
    cache_api_spec,
    API_SPEC_PATH,
)

__all__ = [
    "generate_classes",
    "get_api_spec",
    "generate_classes_maybe_cached",
    "CACHED_SPEC_PATH",
    "cache_api_spec",
    "API_SPEC_PATH",
]



import tempfile
import csv
import gzip
import io
import itertools
import json
import os
import warnings
from functools import partial
from io import StringIO, BytesIO
from unittest import mock
from tempfile import TemporaryDirectory
import zipfile

import pytest
import requests
from requests import ConnectionError, ConnectTimeout

try:
    import pandas as pd

    has_pandas = True
except ImportError:
    has_pandas = False

try:
    import polars as pl

    has_polars = True
except ImportError:
    has_polars = False

import civis
from civis.io import _files
from civis._deprecation import DeprecatedKwargDefault
from civis.io._files import _retry
from civis.io._tables import _File
from civis.io._utils import maybe_get_random_name
from civis.response import Response
from civis.base import CivisAPIError, CivisImportError, EmptyResultError
from civis.tests.mocks import create_client_mock


POLL_INTERVAL = 0.00001

# https://circleci.com/docs/variables/#built-in-environment-variables
ON_CI = os.getenv("CI", "false").lower() == "true"


class MockAPIError(CivisAPIError):
    """A fake API error with only a status code"""

    def __init__(self, sc):
        self.status_code = sc


def _test_file(cols=None, headers=True, delimiter="comma", compression="gzip") -> _File:
    detected_info = {
        "tableColumns": cols,
        "includeHeader": headers,
        "columnDelimiter": delimiter,
        "compression": compression,
    }
    return _File(id=1, name="x.csv", detected_info=detected_info)


@mock.patch.object(_files, "requests", autospec=True)
def test_bytes_file_to_civis(mock_requests):
    mock_civis = create_client_mock()
    civis_name = "somename"
    mock_civis.files.post.return_value.id = 137
    mock_civis.files.post.return_value.url = "url"

    buf = BytesIO()
    buf.write(b"a,b,c\n1,2,3")
    buf.seek(0)

    result = civis.io.file_to_civis(buf, civis_name, client=mock_civis)

    assert isinstance(result, int)


@mock.patch.object(_files, "requests", autospec=True)
def test_zip_member_to_civis(*mocks):
    mock_civis = create_client_mock()
    mock_civis.files.post.return_value.id = 137
    mock_civis.files.post.return_value.url = "url"
    with TemporaryDirectory() as temp_dir:
        fname = os.path.join(temp_dir, "tempfile")
        with zipfile.ZipFile(fname, "w", zipfile.ZIP_DEFLATED) as zip_file:
            archive_name = "archive_name"
            zip_file.writestr(archive_name, "a,b,c\n1,2,3")
            zip_member = zip_file.namelist()[0]
            with zip_file.open(zip_member) as zip_member_buf:
                result = civis.io.file_to_civis(
                    zip_member_buf, zip_member, client=mock_civis
                )

    assert isinstance(result, int)


@mock.patch.object(_files, "requests", autospec=True)
def test_text_file_to_civis(*mocks):
    mock_civis = create_client_mock()
    mock_civis.files.post.return_value.id = 137
    mock_civis.files.post.return_value.url = "url"
    buf = StringIO()
    buf.write("a,b,c\n1,2,3")
    buf.seek(0)
    result = civis.io.file_to_civis(buf, "somename", client=mock_civis)

    assert isinstance(result, int)


@mock.patch.object(_files, "requests", autospec=True)
def test_large_file_to_civis(*mocks):
    mock_civis = create_client_mock()
    mock_civis.files.post_multipart.return_value.id = 137
    mock_civis.files.post_multipart.return_value.upload_urls = ["url"]
    curr_size = civis.io._files.MIN_MULTIPART_SIZE
    civis.io._files.MIN_MULTIPART_SIZE = 1
    with TemporaryDirectory() as temp_dir:
        fname = os.path.join(temp_dir, "tempfile")
        with open(fname, "w+b") as tmp:
            tmp.write(b"a,b,c\n1,2,3")
        with open(fname, "r+b") as tmp:
            result = civis.io.file_to_civis(tmp, fname, client=mock_civis)

        civis.io._files.MIN_MULTIPART_SIZE = curr_size

    assert isinstance(result, int)


@mock.patch("civis.io._tables.file_to_civis")
@mock.patch("civis.io._tables.civis_file_to_table")
def test_csv_to_civis(m_civis_file_to_table, m_file_to_civis):
    mock_civis = create_client_mock()
    mock_file_id = 42
    m_file_to_civis.return_value = mock_file_id
    mock_future = mock.create_autospec(civis.futures.CivisFuture, spec_set=True)
    m_civis_file_to_table.return_value = mock_future
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"

    fname = "a/tempfile"

    with mock.patch.object(
        civis.io._tables,
        "open",
        mock.mock_open(read_data="some,test,data"),
        create=True,
    ) as m_open:

        result = civis.io.csv_to_civis(
            fname, database, table, client=mock_civis, existing_table_rows="truncate"
        )

        m_file_to_civis.assert_called_once_with(
            m_open.return_value, "tempfile", client=mock_civis
        )

    assert result == mock_future

    m_civis_file_to_table.assert_called_once_with(
        mock_file_id,
        database,
        table,
        client=mock_civis,
        max_errors=None,
        existing_table_rows="truncate",
        diststyle=None,
        distkey=None,
        sortkey1=None,
        sortkey2=None,
        table_columns=None,
        delimiter=",",
        headers=None,
        primary_keys=None,
        last_modified_keys=None,
        escaped=False,
        execution="immediate",
        credential_id=None,
        polling_interval=None,
        hidden=True,
    )


@mock.patch.object(_files, "requests", autospec=True)
def test_civis_to_file(mock_requests):
    mock_civis = create_client_mock()
    mock_civis.files.get.return_value.id = 137
    mock_civis.files.get.return_value.upload_url = "url"
    mock_requests.get.return_value.iter_content.return_value = (
        ch.encode() for ch in "a,b,c\n1,2,3"
    )
    buf = BytesIO()
    civis.io.civis_to_file(137, buf, client=mock_civis)
    buf.seek(0)
    assert buf.read() == b"a,b,c\n1,2,3"


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_table_exists(m_run_cleaning, m_process_cleaning_results):
    mock_civis = create_client_mock()
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"
    mock_file_id = 1234
    mock_cleaned_file_id = 1235
    mock_import_id = 8675309

    mock_civis.imports.post_files_csv.return_value.id = mock_import_id
    mock_civis.get_database_id.return_value = 42
    mock_civis.default_database_credential_id = 713

    mock_civis.databases.get_schemas_tables.return_value = Response({"name": "table1"})
    m_process_cleaning_results.return_value = (
        [mock_cleaned_file_id],
        True,  # headers
        "gzip",  # compression
        "comma",  # delimiter
        None,  # table_columns
    )
    m_run_cleaning.return_value = [mock.sentinel.cleaning_future]

    with mock.patch.object(civis.io._tables, "run_job", spec_set=True) as m_run_job:

        run_job_future = mock.MagicMock(
            spec=civis.futures.CivisFuture, job_id=123, run_id=234
        )

        m_run_job.return_value = run_job_future

        result = civis.io.civis_file_to_table(
            mock_file_id,
            database,
            table,
            existing_table_rows="truncate",
            delimiter=",",
            headers=True,
            client=mock_civis,
        )

        assert result is run_job_future
        m_run_job.assert_called_once_with(
            mock_import_id, client=mock_civis, polling_interval=None
        )

    m_run_cleaning.assert_called_once_with(
        [mock_file_id], mock_civis, False, True, "comma", True
    )
    m_process_cleaning_results.assert_called_once_with(
        [mock.sentinel.cleaning_future], mock_civis, True, False, "comma"
    )

    expected_name = "CSV import to scratch.api_client_test_fixture"
    expected_kwargs = {
        "name": expected_name,
        "max_errors": None,
        "existing_table_rows": "truncate",
        "hidden": True,
        "column_delimiter": "comma",
        "compression": "gzip",
        "escaped": False,
        "execution": "immediate",
        "loosen_types": False,
        "table_columns": None,
        "redshift_destination_options": {
            "diststyle": None,
            "distkey": None,
            "sortkeys": [None, None],
        },
    }
    mock_civis.imports.post_files_csv.assert_called_once_with(
        {"file_ids": [mock_cleaned_file_id]},
        {
            "schema": "scratch",
            "table": "api_client_test_fixture",
            "remote_host_id": 42,
            "credential_id": 713,
            "primary_keys": None,
            "last_modified_keys": None,
        },
        True,
        **expected_kwargs,
    )


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_table_doesnt_exist(
    m_run_cleaning, m_process_cleaning_results
):
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"
    mock_file_id = 1234
    mock_cleaned_file_id = 1235
    mock_import_id = 8675309
    mock_civis = create_client_mock()

    mock_civis.imports.post_files_csv.return_value.id = mock_import_id
    mock_civis.get_database_id.return_value = 42
    mock_civis.default_database_credential_id = 713

    mock_civis.databases.get_schemas_tables.side_effect = MockAPIError(404)
    mock_columns = [{"name": "foo", "sql_type": "INTEGER"}]
    m_process_cleaning_results.return_value = (
        [mock_cleaned_file_id],
        True,  # headers
        "gzip",  # compression
        "comma",  # delimiter
        mock_columns,  # table_columns
    )
    m_run_cleaning.return_value = [mock.sentinel.cleaning_future]

    with mock.patch.object(civis.io._tables, "run_job", spec_set=True) as m_run_job:

        run_job_future = mock.MagicMock(
            spec=civis.futures.CivisFuture, job_id=123, run_id=234
        )

        m_run_job.return_value = run_job_future

        result = civis.io.civis_file_to_table(
            mock_file_id,
            database,
            table,
            existing_table_rows="truncate",
            delimiter=",",
            headers=True,
            client=mock_civis,
        )

        assert result is run_job_future
        m_run_job.assert_called_once_with(
            mock_import_id, client=mock_civis, polling_interval=None
        )

    m_run_cleaning.assert_called_once_with(
        [mock_file_id], mock_civis, True, True, "comma", True
    )
    m_process_cleaning_results.assert_called_once_with(
        [mock.sentinel.cleaning_future], mock_civis, True, True, "comma"
    )

    expected_name = "CSV import to scratch.api_client_test_fixture"
    expected_kwargs = {
        "name": expected_name,
        "max_errors": None,
        "existing_table_rows": "truncate",
        "hidden": True,
        "column_delimiter": "comma",
        "compression": "gzip",
        "escaped": False,
        "execution": "immediate",
        "loosen_types": True,
        "table_columns": mock_columns,
        "redshift_destination_options": {
            "diststyle": None,
            "distkey": None,
            "sortkeys": [None, None],
        },
    }
    mock_civis.imports.post_files_csv.assert_called_once_with(
        {"file_ids": [mock_cleaned_file_id]},
        {
            "schema": "scratch",
            "table": "api_client_test_fixture",
            "remote_host_id": 42,
            "credential_id": 713,
            "primary_keys": None,
            "last_modified_keys": None,
        },
        True,
        **expected_kwargs,
    )


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_table_doesnt_exist_all_sql_types_missing(
    m_run_cleaning, m_process_cleaning_results
):
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"
    mock_file_id = 1234
    mock_cleaned_file_id = 1235
    mock_import_id = 8675309
    mock_civis = create_client_mock()

    mock_civis.imports.post_files_csv.return_value.id = mock_import_id
    mock_civis.get_database_id.return_value = 42
    mock_civis.default_database_credential_id = 713
    mock_civis.databases.get_schemas_tables.side_effect = MockAPIError(404)
    table_columns = [{"name": "a", "sql_type": ""}, {"name": "b", "sql_type": ""}]
    detected_columns = [
        {"name": "a", "sql_type": "INTEGER"},
        {"name": "b", "sql_type": "VARCHAR(42)"},
    ]
    m_process_cleaning_results.return_value = (
        [mock_cleaned_file_id],
        True,  # headers
        "gzip",  # compression
        "comma",  # delimiter
        detected_columns,  # table_columns
    )
    m_run_cleaning.return_value = [mock.sentinel.cleaning_future]

    with mock.patch.object(civis.io._tables, "run_job", spec_set=True) as m_run_job:

        run_job_future = mock.MagicMock(
            spec=civis.futures.CivisFuture, job_id=123, run_id=234
        )

        m_run_job.return_value = run_job_future

        result = civis.io.civis_file_to_table(
            mock_file_id,
            database,
            table,
            existing_table_rows="truncate",
            delimiter=",",
            headers=True,
            client=mock_civis,
            table_columns=table_columns,
        )

        assert result is run_job_future
        m_run_job.assert_called_once_with(
            mock_import_id, client=mock_civis, polling_interval=None
        )

    m_run_cleaning.assert_called_once_with(
        [mock_file_id], mock_civis, True, True, "comma", True
    )
    m_process_cleaning_results.assert_called_once_with(
        [mock.sentinel.cleaning_future], mock_civis, True, True, "comma"
    )

    expected_name = "CSV import to scratch.api_client_test_fixture"
    expected_kwargs = {
        "name": expected_name,
        "max_errors": None,
        "existing_table_rows": "truncate",
        "hidden": True,
        "column_delimiter": "comma",
        "compression": "gzip",
        "escaped": False,
        "execution": "immediate",
        "loosen_types": True,
        "table_columns": detected_columns,
        "redshift_destination_options": {
            "diststyle": None,
            "distkey": None,
            "sortkeys": [None, None],
        },
    }
    mock_civis.imports.post_files_csv.assert_called_once_with(
        {"file_ids": [mock_cleaned_file_id]},
        {
            "schema": "scratch",
            "table": "api_client_test_fixture",
            "remote_host_id": 42,
            "credential_id": 713,
            "primary_keys": None,
            "last_modified_keys": None,
        },
        True,
        **expected_kwargs,
    )


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_table_does_not_exist_some_sql_types_missing(
    m_run_cleaning, m_process_cleaning_results
):
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"
    mock_file_id = 1234
    mock_import_id = 8675309
    mock_civis = create_client_mock()

    mock_civis.imports.post_files_csv.return_value.id = mock_import_id
    mock_civis.get_database_id.return_value = 42
    mock_civis.default_database_credential_id = 713
    mock_civis.databases.get_schemas_tables.side_effect = MockAPIError(404)
    table_columns = [{"name": "a", "sql_type": "INT"}, {"name": "b", "sql_type": ""}]

    with pytest.raises(ValueError):
        civis.io.civis_file_to_table(
            mock_file_id,
            database,
            table,
            existing_table_rows="truncate",
            delimiter=",",
            headers=True,
            client=mock_civis,
            table_columns=table_columns,
        )


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_table_columns_keys_misspelled(
    m_run_cleaning, m_process_cleaning_results
):
    # Check for an error message if the `table_columns` input
    # contains keys other than the accepted ones.
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"
    mock_file_id = 1234
    mock_import_id = 8675309
    mock_civis = create_client_mock()

    mock_civis.imports.post_files_csv.return_value.id = mock_import_id
    mock_civis.get_database_id.return_value = 42
    mock_civis.default_database_credential_id = 713
    mock_civis.databases.get_schemas_tables.side_effect = MockAPIError(404)
    table_columns = [{"name": "a", "sqlType": "INT"}, {"name": "b", "bad_type": ""}]

    with pytest.raises(ValueError) as err:
        civis.io.civis_file_to_table(
            mock_file_id,
            database,
            table,
            existing_table_rows="drop",
            delimiter=",",
            headers=True,
            client=mock_civis,
            table_columns=table_columns,
        )
    assert "must be one of ('name', 'sql_type')" in str(err.value)
    assert "also has ('bad_type', 'sqlType')" in str(err.value)


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_table_doesnt_exist_provide_table_columns(
    m_run_cleaning, m_process_cleaning_results
):

    # pytest.parametrize apparently doesn't work with unittest.TestCase
    # objects, which this ultimately inherits from, so we'll define a
    # function here and run subtests below.
    # See https://docs.pytest.org/en/stable/unittest.html.
    def run_subtest(mock_file_ids):
        table = "scratch.api_client_test_fixture"
        database = "redshift-general"
        mock_cleaned_file_ids = mock.Mock()
        mock_import_id = 8675309
        mock_civis = create_client_mock()

        mock_civis.imports.post_files_csv.return_value.id = mock_import_id
        mock_civis.get_database_id.return_value = 42
        mock_civis.default_database_credential_id = 713
        mock_civis.databases.get_schemas_tables.side_effect = MockAPIError(404)
        table_columns = [
            {"name": "foo", "sql_type": "INTEGER"},
            {"name": "bar", "sql_type": "VARCHAR(42)"},
        ]
        m_process_cleaning_results.return_value = (
            mock_cleaned_file_ids,
            True,  # headers
            "gzip",  # compression
            "comma",  # delimiter
            None,  # table_columns
        )
        m_run_cleaning.return_value = [mock.sentinel.cleaning_future]

        with mock.patch.object(civis.io._tables, "run_job", spec_set=True) as m_run_job:

            run_job_future = mock.MagicMock(
                spec=civis.futures.CivisFuture, job_id=123, run_id=234
            )

            m_run_job.return_value = run_job_future

            result = civis.io.civis_file_to_table(
                mock_file_ids,
                database,
                table,
                existing_table_rows="truncate",
                table_columns=table_columns,
                delimiter=",",
                headers=True,
                client=mock_civis,
            )

            assert result is run_job_future
            m_run_job.assert_called_once_with(
                mock_import_id, client=mock_civis, polling_interval=None
            )

        m_run_cleaning.assert_called_once_with(
            [mock_file_ids] if isinstance(mock_file_ids, int) else mock_file_ids,
            mock_civis,
            False,
            True,
            "comma",
            True,
        )
        m_process_cleaning_results.assert_called_once_with(
            [mock.sentinel.cleaning_future], mock_civis, True, False, "comma"
        )

        expected_name = "CSV import to scratch.api_client_test_fixture"
        expected_kwargs = {
            "name": expected_name,
            "max_errors": None,
            "existing_table_rows": "truncate",
            "hidden": True,
            "column_delimiter": "comma",
            "compression": "gzip",
            "escaped": False,
            "execution": "immediate",
            "loosen_types": False,
            "table_columns": table_columns,
            "redshift_destination_options": {
                "diststyle": None,
                "distkey": None,
                "sortkeys": [None, None],
            },
        }
        mock_civis.imports.post_files_csv.assert_called_once_with(
            {"file_ids": mock_cleaned_file_ids},
            {
                "schema": "scratch",
                "table": "api_client_test_fixture",
                "remote_host_id": 42,
                "credential_id": 713,
                "primary_keys": None,
                "last_modified_keys": None,
            },
            True,
            **expected_kwargs,
        )

    # Check that things work with a single file ID or multiple IDs.
    # In particular, we want to ensure that loosen_types is set to False
    # in both situations.
    for mock_file_ids in (1234, [1234], [1234, 1235]):
        m_run_cleaning.reset_mock()
        m_process_cleaning_results.reset_mock()
        run_subtest(mock_file_ids)


@mock.patch("civis.io._tables._process_cleaning_results")
@mock.patch("civis.io._tables._run_cleaning")
def test_civis_file_to_table_multi_file(m_run_cleaning, m_process_cleaning_results):
    table = "scratch.api_client_test_fixture"
    database = "redshift-general"
    mock_file_id = [1234, 1235]
    mock_cleaned_file_ids = [1236, 1237]
    mock_import_id = 8675309
    mock_civis = create_client_mock()

    mock_civis.imports.post_files_csv.return_value.id = mock_import_id
    mock_civis.get_database_id.return_value = 42
    mock_civis.default_database_credential_id = 713

    mock_civis.databases.get_schemas_tables.side_effect = MockAPIError(404)
    mock_columns = [{"name": "foo", "sql_type": "INTEGER"}]
    m_process_cleaning_results.return_value = (
        mock_cleaned_file_ids,
        True,  # headers
        "gzip",  # compression
        "comma",  # delimiter
        mock_columns,  # table_columns
    )
    m_run_cleaning.return_value = [
        mock.sentinel.cleaning_future1,
        mock.sentinel.cleaning_future2,
    ]

    with mock.patch.object(civis.io._tables, "run_job", spec_set=True) as m_run_job:

        run_job_future = mock.MagicMock(
            spec=civis.futures.CivisFuture, job_id=123, run_id=234
        )

        m_run_job.return_value = run_job_future

        result = civis.io.civis_file_to_table(
            mock_file_id,
            database,
            table,
            existing_table_rows="truncate",
            delimiter=",",
            headers=True,
            client=mock_civis,
        )

        assert result is run_job_future
        m_run_job.assert_called_once_with(
            mock_import_id, client=mock_civis, polling_interval=None
        )

    m_run_cleaning.assert_called_once_with(
        mock_file_id, mock_civis, True, True, "comma", True
    )
    m_process_cleaning_results.assert_called_once_with(
        [mock.sentinel.cleaning_future1, mock.sentinel.cleaning_future2],
        mock_civis,
        True,
        True,
        "comma",
    )

    expected_name = "CSV import to scratch.api_client_test_fixture"
    expected_kwargs = {
        "name": expected_name,
        "max_errors": None,
        "existing_table_rows": "truncate",
        "hidden": True,
        "column_delimiter": "comma",
        "compression": "gzip",
        "escaped": False,
        "execution": "immediate",
        "loosen_types": True,
        "table_columns": mock_columns,
        "redshift_destination_options": {
            "diststyle": None,
            "distkey": None,
            "sortkeys": [None, None],
        },
    }
    mock_civis.imports.post_files_csv.assert_called_once_with(
        {"file_ids": mock_cleaned_file_ids},
        {
            "schema": "scratch",
            "table": "api_client_test_fixture",
            "remote_host_id": 42,
            "credential_id": 713,
            "primary_keys": None,
            "last_modified_keys": None,
        },
        True,
        **expected_kwargs,
    )


def test_process_cleaning_results():
    mock_job_id = 42
    mock_run_id = 1776
    mock_file_id = 312
    mock_civis = create_client_mock()
    fut = civis.futures.CivisFuture(
        poller=lambda j, r: (j, r),
        poller_args=(mock_job_id, mock_run_id),
        poll_on_creation=False,
        client=mock_civis,
    )
    fut.set_result(Response({"state": "success"}))

    mock_civis.jobs.list_runs_outputs.return_value = [
        Response({"object_id": mock_file_id})
    ]

    expected_columns = [
        {"name": "a", "sql_type": "INT"},
        {"name": "column", "sql_type": "INT"},
    ]
    expected_compression = "gzip"
    expected_headers = True
    expected_delimiter = ","
    mock_civis.files.get.return_value = Response(
        {
            "id": mock_file_id,
            "detected_info": {
                "tableColumns": expected_columns,
                "compression": expected_compression,
                "includeHeader": expected_headers,
                "columnDelimiter": ",",
            },
            "name": "file1.csv.gz",
        }
    )

    assert civis.io._tables._process_cleaning_results(
        [fut], mock_civis, None, True, None
    ) == (
        [mock_file_id],
        expected_headers,
        expected_compression,
        expected_delimiter,
        expected_columns,
    )


def test_process_cleaning_results_raises_imports():
    mock_job_id = 42
    mock_run_id = 1776
    mock_file_id = 312
    mock_civis = create_client_mock()
    fut = civis.futures.CivisFuture(
        poller=lambda j, r: (j, r),
        poller_args=(mock_job_id, mock_run_id),
        poll_on_creation=False,
        client=mock_civis,
    )
    fut.set_result(Response({"state": "success"}))

    fut2 = civis.futures.CivisFuture(
        poller=lambda j, r: (j, r),
        poller_args=(mock_job_id, mock_run_id),
        poll_on_creation=False,
        client=mock_civis,
    )
    fut2.set_result(Response({"state": "success"}))

    mock_civis.jobs.list_runs_outputs.return_value = [
        Response({"object_id": mock_file_id})
    ]

    expected_compression = "gzip"
    expected_headers = True
    expected_cols = [
        {"name": "a", "sql_type": "INT"},
        {"name": "column", "sql_type": "INT"},
    ]
    resp1 = Response(
        {
            "id": 123,
            "detected_info": {
                "tableColumns": expected_cols,
                "compression": expected_compression,
                "includeHeader": expected_headers,
                "columnDelimiter": ",",
            },
            "name": "file1.csv.gz",
        }
    )

    resp2 = Response(
        {
            "id": 456,
            "detected_info": {
                "tableColumns": expected_cols,
                "compression": expected_compression,
                "includeHeader": expected_headers,
                "columnDelimiter": "|",
            },
            "name": "file2.csv.gz",
        }
    )
    mock_civis.files.get.side_effect = [resp1, resp2]

    regex = (
        r"All detected values for 'columnDelimiter' "
        r"must be the same, however --\n"
        r"\t, from: file 123 \(file1.csv.gz\)\n"
        r"\t| from: file 456 \(file2.csv.gz\)"
    )
    with pytest.raises(CivisImportError, match=regex):
        civis.io._tables._process_cleaning_results(
            [fut, fut2], mock_civis, None, True, None
        )


@pytest.mark.parametrize("fids", ([42], [42, 43]))
@mock.patch("civis.io._tables.run_job")
def test_run_cleaning(m_run_job, fids):

    def mock_preprocess(
        file_id,
        in_place,
        detect_table_columns=True,
        force_character_set_conversion=True,
        include_header=True,
        column_delimiter="comma",
        hidden=True,
    ):
        resp = Response({"id": file_id})
        return resp

    mock_civis = create_client_mock()

    mock_civis.files.post_preprocess_csv.side_effect = mock_preprocess
    mock_future = mock.create_autospec(civis.futures.CivisFuture, spec_set=True)
    m_run_job.return_value = mock_future
    res = civis.io._tables._run_cleaning(fids, mock_civis, True, True, "comma", True)

    # We should have one cleaning job per provided file id
    fid_count = len(fids)
    assert len(res) == fid_count
    mock_civis.files.post_preprocess_csv.assert_has_calls(
        (
            mock.call(
                file_id=fid,
                in_place=False,
                detect_table_columns=True,
                force_character_set_conversion=True,
                include_header=True,
                column_delimiter="comma",
                hidden=True,
            )
            for fid in fids
        )
    )
    m_run_job.assert_has_calls(
        (mock.call(jid, client=mock_civis, polling_interval=None) for jid in fids)
    )


def test_check_detected_info_matching():
    files = [
        _test_file(headers=True, delimiter="comma", compression="gzip"),
        _test_file(headers=True, delimiter="comma", compression="gzip"),
    ]
    for attr in ("includeHeader", "columnDelimiter", "compression"):
        civis.io._tables._check_detected_info(files, attr)


def test_check_detected_info_raises():
    files = [
        _test_file(headers=True, delimiter="comma", compression="gzip"),
        _test_file(headers=False, delimiter="pipe", compression="none"),
    ]
    for attr in ("includeHeader", "columnDelimiter", "compression"):
        with pytest.raises(civis.base.CivisImportError):
            civis.io._tables._check_detected_info(files, attr)


def test_check_column_types_differing_numbers():
    files = [
        _test_file([{"name": "col1", "sql_type": "INT"}]),
        _test_file(
            [{"name": "col1", "sql_type": "INT"}, {"name": "col2", "sql_type": "FLOAT"}]
        ),
    ]
    regex = (
        r"All files must have the same number of columns, however --\n"
        r"\t1 from: file 1 \(x.csv\)\n"
        r"\t2 from: file 1 \(x.csv\)"
    )
    with pytest.raises(civis.base.CivisImportError, match=regex):
        civis.io._tables._check_column_types(files)


def test_check_column_types_differing_types():
    files = [
        _test_file([{"name": "col1", "sql_type": "INT"}]),
        _test_file([{"name": "col1", "sql_type": "FLOAT"}]),
    ]
    regex = (
        r"All sql_types for column 'col1' must be the same, however --\n"
        r"\tINT from: file 1 \(x.csv\)\n"
        r"\tFLOAT from: file 1 \(x.csv\)"
    )
    with pytest.raises(civis.base.CivisImportError, match=regex):
        civis.io._tables._check_column_types(files)


def test_check_column_types_passing():
    files = [
        _test_file(
            [
                {"name": "col1", "sql_type": "INT"},
                {"name": "col2", "sql_type": "VARCHAR(42)"},
            ]
        ),
        _test_file(
            [
                {"name": "col1", "sql_type": "INT"},
                {"name": "col2", "sql_type": "VARCHAR(47)"},
            ]
        ),
    ]
    actual, allow_inconsistent_headers = civis.io._tables._check_column_types(files)
    expected = [
        {"name": "col1", "sql_type": "INT"},
        {"name": "col2", "sql_type": "VARCHAR(42)"},
    ]
    assert actual == expected
    assert allow_inconsistent_headers is False


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
@mock.patch("civis.io._tables.file_to_civis")
@mock.patch("civis.io._tables.civis_file_to_table")
def test_dataframe_to_civis_pandas(m_civis_file_to_table, m_file_to_civis):
    mock_civis = create_client_mock()
    df = pd.DataFrame([[1, 2, 3], [2, 3, 4]])
    m_file_to_civis.return_value = 42
    mock_future = mock.create_autospec(civis.futures.CivisFuture, spec_set=True)
    m_civis_file_to_table.return_value = mock_future

    # use a mock to spy on the dataframe's to_csv method so we can
    # check on its calls without impeding its actual usage.
    with mock.patch.object(df, "to_csv", wraps=df.to_csv) as m_to_csv:
        result = civis.io.dataframe_to_civis(
            df,
            "redshift-general",
            "scratch.api_client_test_fixture",
            existing_table_rows="truncate",
            client=mock_civis,
        )
        assert result == mock_future

        # ANY here represents the path to which the dataframe was written
        # Since it's a temporary directory we don't know/care exactly what
        # it is
        m_to_csv.assert_called_once_with(mock.ANY, encoding="utf-8", index=False)
        out_path = m_to_csv.call_args.args[0]

    m_file_to_civis.assert_called_once_with(
        mock.ANY, "api_client_test_fixture", client=mock_civis
    )

    # Ensure that the file written to above is the same file as that
    # uploaded to Civis in this call
    assert m_file_to_civis.call_args.args[0] == out_path

    m_civis_file_to_table.assert_called_once_with(
        m_file_to_civis.return_value,
        "redshift-general",
        "scratch.api_client_test_fixture",
        client=mock_civis,
        max_errors=None,
        existing_table_rows="truncate",
        diststyle=None,
        distkey=None,
        sortkey1=None,
        sortkey2=None,
        table_columns=None,
        delimiter=",",
        primary_keys=None,
        last_modified_keys=None,
        escaped=False,
        execution="immediate",
        headers=True,
        credential_id=None,
        polling_interval=None,
        hidden=True,
    )


@pytest.mark.skipif(not ON_CI and not has_polars, reason="polars not installed")
@mock.patch("civis.io._tables.file_to_civis")
@mock.patch("civis.io._tables.civis_file_to_table")
def test_dataframe_to_civis_polars(m_civis_file_to_table, m_file_to_civis):
    mock_civis = create_client_mock()
    df = pl.DataFrame([[1, 2, 3], [2, 3, 4]], orient="row")
    m_file_to_civis.return_value = 42
    mock_future = mock.create_autospec(civis.futures.CivisFuture, spec_set=True)
    m_civis_file_to_table.return_value = mock_future

    # use a mock to spy on the dataframe's to_csv method so we can
    # check on its calls without impeding its actual usage.
    with mock.patch.object(df, "write_csv", wraps=df.write_csv) as m_write_csv:
        result = civis.io.dataframe_to_civis(
            df,
            "redshift-general",
            "scratch.api_client_test_fixture",
            existing_table_rows="truncate",
            client=mock_civis,
        )
        assert result == mock_future

        # ANY here represents the path to which the dataframe was written
        # Since it's a temporary directory we don't know/care exactly what
        # it is
        m_write_csv.assert_called_once_with(mock.ANY)
        out_path = m_write_csv.call_args.args[0]

    m_file_to_civis.assert_called_once_with(
        mock.ANY, "api_client_test_fixture", client=mock_civis
    )

    # Ensure that the file written to above is the same file as that
    # uploaded to Civis in this call
    assert m_file_to_civis.call_args.args[0] == out_path

    m_civis_file_to_table.assert_called_once_with(
        m_file_to_civis.return_value,
        "redshift-general",
        "scratch.api_client_test_fixture",
        client=mock_civis,
        max_errors=None,
        existing_table_rows="truncate",
        diststyle=None,
        distkey=None,
        sortkey1=None,
        sortkey2=None,
        table_columns=None,
        delimiter=",",
        primary_keys=None,
        last_modified_keys=None,
        escaped=False,
        execution="immediate",
        headers=True,
        credential_id=None,
        polling_interval=None,
        hidden=True,
    )


def test_check_column_types_coerce_to_varchar():
    case1 = [
        _test_file([{"name": "col1", "sql_type": "INT"}]),
        _test_file([{"name": "col1", "sql_type": "VARCHAR(42)"}]),
    ]
    case2 = [
        _test_file([{"name": "col1", "sql_type": "VARCHAR(42)"}]),
        _test_file([{"name": "col1", "sql_type": "INT"}]),
    ]
    case3 = [
        _test_file([{"name": "col1", "sql_type": "INT"}]),
        _test_file([{"name": "col1", "sql_type": "VARCHAR(42)"}]),
        _test_file([{"name": "col1", "sql_type": "FLOAT"}]),
    ]
    case4 = [
        _test_file([{"name": "col1", "sql_type": "INT"}]),
        _test_file([{"name": "col1", "sql_type": "FLOAT"}]),
        _test_file([{"name": "col1", "sql_type": "VARCHAR(42)"}]),
    ]
    case5 = [
        _test_file([{"name": "col1", "sql_type": "INT"}]),
        _test_file([{"name": "col1", "sql_type": "VARCHAR(42)"}]),
        _test_file([{"name": "col1", "sql_type": "FLOAT"}]),
        _test_file([{"name": "col1", "sql_type": "VARCHAR(8)"}]),
    ]
    for files in (case1, case2, case3, case4, case5):
        actual, allow_inconsistent_headers = civis.io._tables._check_column_types(files)
        expected = [{"name": "col1", "sql_type": "VARCHAR"}]
        assert actual == expected, f"failed for {files}"
        assert allow_inconsistent_headers is True


@mock.patch("civis.io._tables.CivisFuture")
@mock.patch("civis.io._tables.civis_to_file")
@mock.patch("civis.io._tables._sql_script")
def test_civis_to_multifile_passes_client(
    m_sql_script, m_civis_to_file, m_CivisFuture, *mocks
):
    """Ensure the client kwarg is passed forward."""
    m_sql_script.return_value = (mock.MagicMock(), mock.MagicMock())
    # We need to write some JSON into the buffer to avoid errors.
    m_civis_to_file.side_effect = lambda _, buf, *args, **kwargs: buf.write(b"{}")
    mock_client = mock.MagicMock()

    civis.io.civis_to_multifile_csv("sql", "db", client=mock_client)

    m_civis_to_file.assert_called_once_with(mock.ANY, mock.ANY, client=mock_client)


@mock.patch("civis.io.query_civis")
@mock.patch("civis.io.transfer_table")
def test_transfer_table(m_transfer_table, m_query_civis):
    mock_civis = create_client_mock()
    mock_future = mock.create_autospec(civis.futures.CivisFuture)
    resp = Response(
        {"state": "succeeded"},
    )
    attrs = {"result.return_value": resp, "state": "succeeded"}
    mock_future.configure_mock(**attrs)
    m_transfer_table.return_value = mock_future
    result = civis.io.transfer_table(
        "redshift-general",
        "redshift-test",
        "scratch.api_client_test_fixture",
        "scratch.api_client_test_fixture",
        polling_interval=POLL_INTERVAL,
        client=mock_civis,
    )
    result = result.result()
    assert result.state == "succeeded"

    # check for side effect
    m_query_civis.return_value = mock_future
    sql = "select * from scratch.api_client_test_fixture"
    result = civis.io.query_civis(
        sql, "redshift-test", polling_interval=POLL_INTERVAL, client=mock_civis
    ).result()
    assert result.state == "succeeded"


@mock.patch("civis.io._tables._download_file")
def test_download_file(m_download_file):
    expected = '"1","2","3"\n'
    m_download_file.return_value = expected
    with TemporaryDirectory() as temp_dir:
        fname = os.path.join(temp_dir, "tempfile")
        with open(fname, "w") as tmp:
            tmp.write(expected)
        civis.io._tables._download_file("test_url", fname, b"", "none")
        with open(fname, "r") as f:
            data = f.read()
    assert data == expected


def test_get_sql_select(*mocks):
    x = "select * from schema.table"
    y = "select a, b, c from schema.table"
    table = "schema.table"
    assert civis.io._tables._get_sql_select(table) == x
    assert civis.io._tables._get_sql_select(table, ["a", "b", "c"]) == y
    with pytest.raises(TypeError):
        civis.io._tables._get_sql_select(table, "column_a")


def test_file_id_from_run_output_exact():
    m_client = mock.Mock()
    m_client.jobs.list_runs_outputs.return_value = [
        Response({"name": "spam", "object_id": 2013, "object_type": "File"})
    ]

    fid = civis.io.file_id_from_run_output("spam", 17, 13, client=m_client)
    assert fid == 2013


def test_file_id_from_run_output_approximate():
    # Test fuzzy name matching
    m_client = mock.Mock()
    m_client.jobs.list_runs_outputs.return_value = [
        Response({"name": "spam.csv.gz", "object_id": 2013, "object_type": "File"})
    ]

    fid = civis.io.file_id_from_run_output("^spam", 17, 13, regex=True, client=m_client)
    assert fid == 2013


def test_file_id_from_run_output_approximate_multiple():
    # Fuzzy name matching with muliple matches should return the first
    m_cl = mock.Mock()
    m_cl.jobs.list_runs_outputs.return_value = [
        Response({"name": "spam.csv.gz", "object_id": 2013, "object_type": "File"}),
        Response({"name": "eggs.csv.gz", "object_id": 2014, "object_type": "File"}),
    ]

    fid = civis.io.file_id_from_run_output(".csv", 17, 13, regex=True, client=m_cl)
    assert fid == 2013


def test_file_id_from_run_output_no_file():
    # Get an IOError if we request a file which doesn't exist
    m_client = mock.Mock()
    m_client.jobs.list_runs_outputs.return_value = [
        Response({"name": "spam", "object_id": 2013, "object_type": "File"})
    ]

    with pytest.raises(FileNotFoundError) as err:
        civis.io.file_id_from_run_output("eggs", 17, 13, client=m_client)
    assert "not an output" in str(err.value)


def test_file_id_from_run_output_no_run():
    # Get an IOError if we request a file from a run which doesn't exist
    m_client = mock.Mock()
    m_client.jobs.list_runs_outputs.side_effect = MockAPIError(
        404
    )  # Mock a run which doesn't exist

    with pytest.raises(IOError) as err:
        civis.io.file_id_from_run_output("name", 17, 13, client=m_client)
    assert "could not find job/run id 17/13" in str(err.value).lower()


def test_file_id_from_run_output_platform_error():
    # Make sure we don't swallow real Platform errors
    m_client = mock.Mock()
    m_client.jobs.list_runs_outputs.side_effect = MockAPIError(
        500
    )  # Mock a platform error
    with pytest.raises(CivisAPIError):
        civis.io.file_id_from_run_output("name", 17, 13, client=m_client)


def test_file_id_from_run_output_no_filename():
    m_client = mock.Mock()
    m_client.jobs.list_runs_outputs.return_value = [
        Response({"name": "spam.csv.gz", "object_id": 2013, "object_type": "File"}),
        Response({"name": "eggs.csv.gz", "object_id": 2014, "object_type": "File"}),
    ]

    fid = civis.io.file_id_from_run_output(".*?", 17, 13, regex=True, client=m_client)
    assert fid == 2013


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
def test_file_to_dataframe_expired():
    m_client = mock.Mock()
    url = None
    m_client.files.get.return_value = Response({"name": "spam.csv", "file_url": url})
    expected_err = (
        "Unable to locate file 121. If it previously " + "existed, it may have expired."
    )
    with pytest.raises(EmptyResultError, match=expected_err):
        civis.io.file_to_dataframe(121, client=m_client)


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
def test_file_to_dataframe_infer():
    m_client = mock.Mock()
    url = "url"
    m_client.files.get.return_value = Response({"name": "spam.csv", "file_url": url})
    with mock.patch.object(
        civis.io._files.pd, "read_csv", autospec=True
    ) as mock_read_csv:
        civis.io.file_to_dataframe(121, compression="infer", client=m_client)
        mock_read_csv.assert_called_once_with(url, compression="infer")


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
def test_file_to_dataframe_infer_gzip():
    m_client = mock.Mock()
    url = "url"
    m_client.files.get.return_value = Response({"name": "spam.csv.gz", "file_url": url})
    with mock.patch.object(
        civis.io._files.pd, "read_csv", autospec=True
    ) as mock_read_csv:
        civis.io.file_to_dataframe(121, compression="infer", client=m_client)
        mock_read_csv.assert_called_once_with(url, compression="gzip")


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
def test_file_to_dataframe_kwargs():
    m_client = mock.Mock()
    url = "url"
    m_client.files.get.return_value = Response({"name": "spam.csv", "file_url": url})
    with mock.patch.object(
        civis.io._files.pd, "read_csv", autospec=True
    ) as mock_read_csv:
        civis.io.file_to_dataframe(
            121, compression="special", client=m_client, delimiter="|", nrows=10
        )
        mock_read_csv.assert_called_once_with(
            url, compression="special", delimiter="|", nrows=10
        )


@pytest.mark.skipif(not ON_CI and not has_polars, reason="polars not installed")
def test_file_to_dataframe_polars():
    m_client = mock.Mock()
    url = "url"
    m_client.files.get.return_value = Response({"name": "spam.csv", "file_url": url})
    with mock.patch.object(
        civis.io._files.pl, "read_csv", autospec=True
    ) as mock_read_csv:
        civis.io.file_to_dataframe(
            121, return_as="polars", client=m_client, separator="|", n_rows=10
        )
        mock_read_csv.assert_called_once_with(url, separator="|", n_rows=10)


@mock.patch.object(civis.io._files, "civis_to_file", autospec=True)
def test_load_json(mock_c2f):
    obj = {"spam": "eggs"}

    def _dump_json(file_id, buf, *args, **kwargs):
        buf.write(json.dumps(obj).encode())

    mock_c2f.side_effect = _dump_json
    out = civis.io.file_to_json(13, client=mock.Mock())
    assert out == obj


@mock.patch.object(_files, "requests", autospec=True)
def test_civis_to_file_local(mock_requests):
    # Test that a call to civis_to_file uses `requests` to grab the contents
    # of a URL given by the API client and writes it to a file.
    mock_civis = create_client_mock()
    mock_requests.get.return_value.iter_content.return_value = (
        ch.encode() for ch in "abcdef"
    )
    with TemporaryDirectory() as tdir:
        fname = os.path.join(tdir, "testfile")
        _files.civis_to_file(137, fname, client=mock_civis)
        with open(fname, "rt") as _fin:
            assert _fin.read() == "abcdef"
    mock_civis.files.get.assert_called_once_with(137)
    mock_requests.get.assert_called_once_with(
        mock_civis.files.get.return_value.file_url, stream=True, timeout=60
    )


@mock.patch.object(_files, "requests", autospec=True)
def test_civis_to_file_retries(mock_requests):
    mock_civis = create_client_mock()

    first_try = True

    # Mock the request iter_content so it fails partway the first time.
    def mock_iter_content(_):
        nonlocal first_try
        chunks = [ch.encode() for ch in "abcdef"]
        for i, chunk in enumerate(chunks):

            # Fail partway through on the first try.
            if first_try and i == 3:
                first_try = False
                raise requests.ConnectionError()

            yield chunk

    mock_requests.get.return_value.iter_content = mock_iter_content

    # Add some data to the buffer to test that we seek to the right place
    # when retrying.
    buf = io.BytesIO(b"0123")
    buf.seek(4)

    _files.civis_to_file(137, buf, client=mock_civis)

    # Check that retries work and that the buffer position is reset.
    # If we didn't seek when retrying, we'd get abcabcdef.
    # If we seek'd to position 0, then we'd get abcdef.
    buf.seek(0)
    assert buf.read() == b"0123abcdef"

    mock_civis.files.get.assert_called_once_with(137)
    assert mock_requests.get.call_count == 2
    mock_requests.get.assert_called_with(
        mock_civis.files.get.return_value.file_url, stream=True, timeout=60
    )


@pytest.mark.parametrize("input_filename", ["newname", None])
@mock.patch.object(_files, "requests", autospec=True)
def test_file_to_civis(mock_requests, input_filename):
    # Test that file_to_civis posts a Civis File with the API client
    # and calls `requests.post` on the returned URL.
    mock_civis = create_client_mock()
    civis_name, expected_id = "newname", 137
    mock_civis.files.post.return_value.id = expected_id
    with TemporaryDirectory() as tdir:
        fname = os.path.join(tdir, "newname")
        with open(fname, "wt") as _fout:
            _fout.write("abcdef")
        fid = _files.file_to_civis(
            fname, input_filename, expires_at=None, client=mock_civis
        )
    assert fid == expected_id
    mock_civis.files.post.assert_called_once_with(civis_name, expires_at=None)
    mock_requests.post.assert_called_once_with(
        mock_civis.files.post.return_value.upload_url, files=mock.ANY, timeout=60
    )


def test_file_to_civis_error_for_description_too_long():
    with TemporaryDirectory() as temp_dir:
        file_path = os.path.join(temp_dir, "some_data")
        with open(file_path, "wb") as f:
            f.write(b"foobar")
        with pytest.raises(ValueError, match="longer than 512 characters"):
            _files.file_to_civis(
                file_path, client=create_client_mock(), description="a" * 513
            )


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
@pytest.mark.parametrize(
    "func, should_add_description",
    itertools.product(
        [
            partial(_files.file_to_civis, io.BytesIO(b"some_data"), name="abc"),
            partial(_files.dataframe_to_file, pd.DataFrame({"a": [1]}), name="abc"),
            partial(_files.json_to_file, {"a": 1}, name="abc"),
        ],
        (True, False),
    ),
)
@mock.patch.object(_files, "requests", autospec=True)
def test_file_description_attribute_added_or_not(
    mock_requests, func, should_add_description
):
    mock_client = create_client_mock()
    description = "some_description"
    if should_add_description:
        func(description=description, client=mock_client)
        mock_client.files.post.assert_called_with("abc", description=description)
    else:
        func(client=mock_client)
        mock_client.files.post.assert_called_with("abc")


@pytest.mark.parametrize(
    "table,expected",
    [
        ("schema.table", ("schema", "table")),
        ('schema."t.able"', ("schema", "t.able")),
        ('schema.table"', ("schema", 'table"')),
        ('"sch.ema"."t.able"', ("sch.ema", "t.able")),
        ('schema."tab""le."', ("schema", 'tab"le.')),
        ("table_with_no_schema", (None, "table_with_no_schema")),
    ],
)
def test_split_schema_tablename(table, expected):
    assert civis.io._tables.split_schema_tablename(table) == expected


def test_split_schema_tablename_raises():
    s = "table.with.too.many.periods"
    with pytest.raises(ValueError):
        civis.io._tables.split_schema_tablename(s)


@mock.patch.object(
    civis.io._tables, "_sql_script", autospec=True, return_value=[700, 1000]
)
def test_export_to_civis_file(mock_sql_script):
    expected = [{"file_id": 9844453}]

    mock_client = create_client_mock()
    response = Response({"state": "success", "output": expected})
    mock_client.scripts.get_sql_runs.return_value = response
    mock_client.scripts.post_sql

    sql = "SELECT 1"
    fut = civis.io.export_to_civis_file(
        sql, "fake-db", polling_interval=POLL_INTERVAL, client=mock_client
    )
    data = fut.result()["output"]
    assert data == expected
    mock_sql_script.assert_called_once_with(
        client=mock_client,
        sql=sql,
        database="fake-db",
        job_name=None,
        credential_id=None,
        csv_settings=None,
        hidden=True,
        sql_params_arguments=None,
    )


def test_sql_script():
    sql = "SELECT SPECIAL SQL QUERY"
    export_job_id = 32
    database_id = 29
    credential_id = 3920
    response = Response({"id": export_job_id})

    mock_client = create_client_mock()
    mock_client.scripts.post_sql.return_value = response
    mock_client.get_database_id.return_value = database_id
    mock_client.default_database_credential_id = credential_id

    civis.io._tables._sql_script(
        client=mock_client,
        sql=sql,
        database="fake-db",
        job_name="My job",
        credential_id=None,
        hidden=False,
        csv_settings=None,
    )
    mock_client.scripts.post_sql.assert_called_once_with(
        "My job",
        remote_host_id=database_id,
        credential_id=credential_id,
        sql=sql,
        hidden=False,
        csv_settings={},
    )
    mock_client.scripts.post_sql_runs.assert_called_once_with(export_job_id)


@mock.patch.object(civis.io._tables, "requests")
def test_read_civis_sql_no_dataframe_special_encoding_sad_path(m_requests):
    # Set up a mock client object for what civis.io.read_civis_sql needs.
    m_client = create_client_mock()
    m_client.scripts.get_sql_runs.return_value = Response(
        {
            "output": [{"path": "blah", "file_id": 123, "output_name": "blah"}],
            "state": "success",
        }
    )
    m_response = mock.Mock()
    m_response.raw.read.return_value = gzip.compress(
        # Intentionally using non-ASCII chars like è to mess with encoding
        "foo,bar\n123,très bien\n".encode("latin-1")
    )
    m_requests.get.return_value = m_response

    # The data was encoded in latin-1, but we don't specify this at
    # the `encoding` param in the read_civis_sql call,
    # which should raise UnicodeDecodeError.
    with pytest.raises(UnicodeDecodeError):
        civis.io.read_civis_sql(
            "select 1",
            "db",
            return_as="list",
            client=m_client,
            polling_interval=POLL_INTERVAL,
        )


@mock.patch.object(civis.io._tables, "requests")
def test_read_civis_sql_no_dataframe_special_encoding_happy_path(m_requests):
    # Set up a mock client object for what civis.io.read_civis_sql needs.
    m_client = create_client_mock()
    m_client.scripts.get_sql_runs.return_value = Response(
        {
            "output": [{"path": "blah", "file_id": 123, "output_name": "blah"}],
            "state": "success",
        }
    )

    # Intentionally using non-ASCII chars like è to mess with encoding
    expected_data = "foo,bar\n123,très bien\n"
    encoding = "latin-1"

    m_response = mock.Mock()

    # The helper function _decompress_stream() calls
    # response.raw.read(CHUNK_SIZE) in a while loop.
    # In the mock `m_response.raw.read` here, we use side_effect so that
    # we get the data back from the first call and no data in the second call
    # inside the while loop (because the data has all been consumed in the
    # first call).
    m_response.raw.read.side_effect = [
        gzip.compress(expected_data.encode(encoding)),
        b"",
    ]
    m_requests.get.return_value = m_response

    actual_data = civis.io.read_civis_sql(
        "select 1",
        "db",
        return_as="list",
        client=m_client,
        polling_interval=POLL_INTERVAL,
        encoding=encoding,
    )
    assert list(csv.reader(io.StringIO(expected_data))) == actual_data


@mock.patch.object(civis.io._tables, "requests")
def test_read_civis_sql_no_dataframe(m_requests):
    # Set up a mock client object for what civis.io.read_civis_sql needs.
    m_client = create_client_mock()
    m_client.scripts.get_sql_runs.return_value = Response(
        {
            "output": [{"path": "blah", "file_id": 123, "output_name": "blah"}],
            "state": "success",
        }
    )

    expected_data = "foo,bar\n123,very good\n"

    m_response = mock.Mock()
    # The helper function _decompress_stream() calls
    # response.raw.read(CHUNK_SIZE) in a while loop.
    # In the mock `m_response.raw.read` here, we use side_effect so that
    # we get the data back from the first call and no data in the second call
    # inside the while loop (because the data has all been consumed in the
    # first call).
    m_response.raw.read.side_effect = [
        gzip.compress(expected_data.encode("utf-8")),
        b"",
    ]
    m_requests.get.return_value = m_response

    actual_data = civis.io.read_civis_sql(
        "select 1",
        "db",
        return_as="list",
        client=m_client,
        polling_interval=POLL_INTERVAL,
    )
    assert list(csv.reader(io.StringIO(expected_data))) == actual_data


@pytest.mark.skipif(not ON_CI and not has_pandas, reason="pandas not installed")
def test_read_civis_sql_pandas():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_csv_path = os.path.join(tmp_dir, "data.csv")
        expected_data = "foo,bar\n123,very good\n"
        with open(tmp_csv_path, "wb") as f:
            f.write(gzip.compress(expected_data.encode("utf-8")))

        # Set up a mock client object for what civis.io.read_civis_sql needs.
        m_client = create_client_mock()
        m_client.scripts.get_sql_runs.return_value = Response(
            {
                "output": [
                    {"path": tmp_csv_path, "file_id": 123, "output_name": "blah"}
                ],
                "state": "success",
            }
        )
        actual_data = civis.io.read_civis_sql(
            "select 1",
            "db",
            return_as="pandas",
            client=m_client,
            polling_interval=POLL_INTERVAL,
        )
        assert pd.read_csv(io.StringIO(expected_data)).equals(actual_data)


@pytest.mark.skipif(not ON_CI and not has_polars, reason="polars not installed")
def test_read_civis_sql_polars():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_csv_path = os.path.join(tmp_dir, "data.csv")
        expected_data = "foo,bar\n123,very good\n"
        with open(tmp_csv_path, "wb") as f:
            f.write(gzip.compress(expected_data.encode("utf-8")))

        # Set up a mock client object for what civis.io.read_civis_sql needs.
        m_client = create_client_mock()
        m_client.scripts.get_sql_runs.return_value = Response(
            {
                "output": [
                    {"path": tmp_csv_path, "file_id": 123, "output_name": "blah"}
                ],
                "state": "success",
            }
        )
        actual_data = civis.io.read_civis_sql(
            "select 1",
            "db",
            return_as="polars",
            client=m_client,
            polling_interval=POLL_INTERVAL,
        )
        assert pl.read_csv(io.StringIO(expected_data)).equals(actual_data)


def test_io_no_retry():
    @_retry(ConnectionError, retries=4, delay=0.1)
    def succeeds():
        counter["i"] += 1
        return "success"

    counter = dict(i=0)
    test_result = succeeds()

    assert test_result == "success"
    assert counter["i"] == 1


def test_io_retry_once():
    @_retry(ConnectionError, retries=4, delay=0.1)
    def fails_once():
        counter["i"] += 1
        if counter["i"] < 2:
            raise ConnectionError("failed")
        else:
            return "success"

    counter = dict(i=0)
    test_result = fails_once()

    assert test_result == "success"
    assert counter["i"] == 2


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_io_retry_limit_reached(m_sleep):
    @_retry(ConnectionError, retries=4, delay=0.1)
    def always_fails():
        counter["i"] += 1
        raise ConnectionError("failed")

    counter = dict(i=0)
    pytest.raises(ConnectionError, always_fails)
    assert counter["i"] == 5


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_io_retry_multiple_exceptions(m_sleep):
    @_retry((ConnectionError, ConnectTimeout), retries=4, delay=0.1)
    def raise_multiple_exceptions():
        counter["i"] += 1
        if counter["i"] == 1:
            raise ConnectionError("one error")
        elif counter["i"] == 2:
            raise requests.ConnectTimeout("another error")
        else:
            return "success"

    counter = dict(i=0)
    test_result = raise_multiple_exceptions()

    assert test_result == "success"
    assert counter["i"] == 3


def test_io_retry_unexpected_exception():
    @_retry(ConnectionError, retries=4, delay=0.1)
    def raise_unexpected_error():
        raise ValueError("unexpected error")

    pytest.raises(ValueError, raise_unexpected_error)


@mock.patch("civis.io._utils.uuid")
def test_maybe_random_name_random(mock_uuid):
    random_name = "11111"
    mock_uuid.uuid4.return_value = mock.Mock(hex=random_name)
    assert maybe_get_random_name(None) == random_name


def test_maybe_random_name_not_random():
    given_name = "22222"
    assert maybe_get_random_name(given_name) == given_name


@pytest.mark.skipif(
    not ON_CI and not (has_pandas and has_polars),
    reason="pandas or polars not installed",
)
@pytest.mark.parametrize(
    "func_name, use_pandas, return_as",
    itertools.product(
        ("read_civis_sql", "read_civis"),
        (True, False, DeprecatedKwargDefault()),
        ("list", "pandas", "polars"),
    ),
)
@mock.patch.object(civis.io._tables, "requests")
def test_warns_or_raise_exception_for_deprecated_use_pandas(
    m_requests, func_name, use_pandas, return_as
):
    if isinstance(use_pandas, DeprecatedKwargDefault):
        warn_or_raise = None
    elif use_pandas is True and return_as == "polars":
        warn_or_raise = ValueError
    elif use_pandas is False and return_as == "pandas":
        warn_or_raise = ValueError
    else:
        warn_or_raise = FutureWarning

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_csv_path = os.path.join(tmp_dir, "data.csv")
        expected_data = "foo,bar\n123,very good\n"
        with open(tmp_csv_path, "wb") as f:
            f.write(gzip.compress(expected_data.encode("utf-8")))

        # Set up a mock client object for what civis.io.read_civis(_sql) needs.
        m_client = create_client_mock()
        m_client.scripts.get_sql_runs.return_value = Response(
            {
                "output": [
                    {"path": tmp_csv_path, "file_id": 123, "output_name": "blah"}
                ],
                "state": "success",
            }
        )
        m_response = mock.Mock()
        # The helper function _decompress_stream() calls
        # response.raw.read(CHUNK_SIZE) in a while loop.
        # In the mock `m_response.raw.read` here, we use side_effect so that
        # we get the data back from the first call and no data in the second call
        # inside the while loop (because the data has all been consumed in the
        # first call).
        m_response.raw.read.side_effect = [
            gzip.compress(expected_data.encode("utf-8")),
            b"",
        ]
        m_requests.get.return_value = m_response

        func = getattr(civis.io, func_name)
        shared_args = dict(
            database="db",
            use_pandas=use_pandas,
            return_as=return_as,
            client=m_client,
            polling_interval=POLL_INTERVAL,
        )
        if func_name == "read_civis":
            args = {"table": "schema.tablename", **shared_args}
        else:
            args = {"sql": "select * from schema.tablename", **shared_args}

        if warn_or_raise and warn_or_raise.__base__.__name__ == "Warning":
            with pytest.warns(warn_or_raise):
                func(**args)
        elif warn_or_raise and warn_or_raise.__base__.__name__ == "Exception":
            with pytest.raises(warn_or_raise):
                func(**args)
        else:
            # Check that no warning is emitted.
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                func(**args)
